/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: camera.h,v 1.1.1.1 2003/07/25 19:28:43 rodzilla Exp $

    The Camera class defines a basic camera.
*/

#ifndef __SAM3D_CAMERA_H__
#define __SAM3D_CAMERA_H__

#include <SaM3d/config.h>
#include <SaM3d/math.h>

namespace SaM3d {

	//!	A simple camera class.
	/*	This basic camera class provides the following features :
		- it can be activated (i.e. the scene will be seen from the
		  camera's point of view)
		- it can be moved, rotated.
		- one can obtain a forward, up and right vector (could
		  be useful to move the camera, to get useful vectors for
		  billboards...).
		\todo The camera should control the field of view angle,
		      and the near and far planes.
	*/
	class Camera {
		//!	A pointer to the currently active camera.
		static Camera*	m_current;

	protected:
		//!	Position of the camera.
		Vector3		m_position;
		//!	Orientation of the camera.
		Quaternion	m_orientation;
		//!	Near plane distance.
		float		m_near;
		//	Far plane distance.
		float		m_far;
		//	View angle.
		float		m_angle;

		//!	A vector facing forward.
		Vector3		m_forward;
		//!	A vector facing up.
		Vector3		m_up;
		//!	A vector facing to the right.
		Vector3		m_right;

	public:
		//!	Default constructor.
		Camera();

		//!	Activate the camera.
		/*!	Change the projection and model view matrices so the
			world is viewed from the camera's point of view.
		*/
		void activate();

		//!	Get the camera's position.
		/*!	\return the camera's position.
		*/
		const Vector3& get_position() const { return m_position; }

		//!	Set the camera's position.
		/*!	\param p the new position.
		*/
		void set_position(const Vector3& p);

		//!	Set the camera's field of view.
		/*!	\param n the distance of the near plane.
			\param f the distance of the far plane.
		*/
		void set_fov(const float n, const float f);

		//!	Set the camera's view angle.
		/*!	\param angle the view angle.
		*/
		void set_angle(const float angle);

		//!	Move the camera.
		/*!	\param v the translation vector in world coordinates.
		*/
		void move(const Vector3& v);

		//!	Rotate the camera.
		/*!	\param q the rotation quaternion in world coordinates.
		*/
		void rotate(const Quaternion& q);

		//!	Return the camera's position.
		/*!	\return the camera's position in world coordinates.
		*/
		const Vector3&  position() const { return m_position; }

		//!	Return the camera's orientation.
		/*!	\return the camera's orientation in world coordinates.
		*/
		const Quaternion& orientation() const { return m_orientation; }

		//!	Return the camera's forward vector.
		/*!	\return the camera's forward vector in world coordinates.
		*/
		const Vector3& forward() const { return m_forward; }

		//!	Return the camera's up vector.
		/*!	\return the camera's up vector in world coordinates.
		*/
		const Vector3& up() const { return m_up; }

		//!	Return the camera's right vector.
		/*!	\return the camera's right vector in world coordinates.
		*/
		const Vector3& right() const { return m_right; }

		//!	Return the current camera.
		static Camera* current() { return m_current; }
	};

}

#endif
