/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2002 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: config_variable.h,v 1.5 2003/09/03 11:13:45 rodzilla Exp $
*/

#ifndef __SAM3D_CONFIG_VARIABLE_H__
#define __SAM3D_CONFIG_VARIABLE_H__

#include <unistd.h>
#include <string>

namespace SaM3d {

	/*!	Base class for configuration variables.
	*/
	class CfgVar {
	private:
		CfgVar();

	protected:
		//!	Variable name.
		const char* m_name;

	public:
		//!	Constructor.
		/*!	\param name the variable name.
		*/
		CfgVar(const char* name);

		//!	Destructor.
		virtual ~CfgVar();

		//!	Find a configuration variable from its name.
		static CfgVar* get(const char* name);

		//!	Get the variable's value as a boolean.
		/*!	\return true if the variable is a boolean variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(bool& value) { return false; }

		//!	Set the variable's value as a boolean.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a boolean variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(bool& value) { return false; }

		//!	Get the variable's value as a float.
		/*!	\return true if the variable is a float variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(float& value) { return false; }

		//!	Set the variable's value as a float.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a float variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(float& value) { return false; }

		//!	Get the variable's value as a string.
		/*!	\return true if the variable is a string variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(std::string& value) { return false; }

		//!	Set the variable's value as a float.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a float variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(std::string& value) { return false; }

	public:
		//!	Base class for float configuration variable handlers.
		struct NumHandler {
			//!	Method to be invoked when assigning a configuration variable.
			/*!	This method is invoked whenever a floar configuration variable
				is set.
				It must be overloaded with useful implementations (that
				check or modify the assigned value, invoke functions...).
				\return true is the assignment must be successful, false otherwise.
				\param value the value to be assigned.
			*/
			virtual bool operator ()(float& value) const { return true; }
		};
	};

	//!	Boolean configuration variable.
	class BoolCfgVar : public CfgVar {
		//!	The variable's value.
		bool m_value;

		//!	The variable's handler.
		bool (*m_handler)(bool&);

		BoolCfgVar();

	public:
		//!	Constructor.
		/*!	This creates a boolean configuration variable and registers it
			into the variable table.
			\param name the variable's name.
			\param value the variable's initial value.
			\param handler the variable's handler.
		*/
		BoolCfgVar(const char* name,
			   const bool value,
			   bool (*handler)(bool&) = NULL);

		//!	Destructor.
		virtual ~BoolCfgVar();

		//!	Bool conversion operator.
		/*!	\return the variable's value.
		*/
		operator bool() const { return m_value; }

		//!	Assignment operator.
		/*!	This operator invokes the set(bool& value) method.
			\return this.
			\param src the value to be assigned.
		*/
		BoolCfgVar& operator=(const bool src);

		//!	Get the variable's value as a boolean.
		/*!	\return true if the variable is a boolean variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(bool& value);

		//!	Set the variable's value as a boolean.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a boolean variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(bool& value);
	};
	
	class IntCfgVar : public CfgVar {
	private:
		//!	The variable's value.
		int m_value;

		//!	The variable's handler.
		NumHandler& m_handler;

		IntCfgVar();

	public:
		//!	Constructor.
		/*!	This creates an integer configuration variable and registers it
			into the variable table.
			\param name the variable's name.
			\param value the variable's initial value.
		*/
		IntCfgVar(const char* name, const int value);

		//!	Constructor.
		/*!	This creates an integer configuration variable and registers it
			into the variable table.
			\param name the variable's name.
			\param value the variable's initial value.
			\param handler the variable's handler.
		*/
		IntCfgVar(const char* name, const int value, NumHandler& handler);

		//!	Destructor.
		virtual ~IntCfgVar();

		//!	Tnteger conversion operator.
		/*!	\return the variable's value.
		*/
		operator int() const { return m_value; }

		//!	Assignment operator.
		/*!	This operator invokes the set(float& value) method.
			\return this.
			\param src the value to be assigned.
		*/
		IntCfgVar& operator=(const int src);

		//!	Get the variable's value as a float.
		/*!	\return true if the variable is a float variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(float& value);

		//!	Set the variable's value as a float.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a float variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(float& value);
	};
	
	class FloatCfgVar : public CfgVar {
	private:
		//!	The variable's value.
		float m_value;

		//!	The variable's handler.
		NumHandler& m_handler;

		FloatCfgVar();

	public:
		//!	Constructor.
		/*!	This creates a float configuration variable and registers it
			into the variable table.
			\param name the variable's name.
			\param value the variable's initial value.
		*/
		FloatCfgVar(const char* name, const float value);

		//!	Constructor.
		/*!	This creates a float configuration variable and registers it
			into the variable table.
			\param name the variable's name.
			\param value the variable's initial value.
			\param handler the variable's handler.
		*/
		FloatCfgVar(const char* name, const float value, NumHandler& handler);

		//!	Destructor.
		virtual ~FloatCfgVar();

		//!	Float conversion operator.
		/*!	\return the variable's value.
		*/
		operator float() const { return m_value; }

		//!	Assignment operator.
		/*!	This operator invokes the set(float& value) method.
			\return this.
			\param src the value to be assigned.
		*/
		FloatCfgVar& operator=(const float src);

		//!	Get the variable's value as a float.
		/*!	\return true if the variable is a float variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(float& value);

		//!	Set the variable's value as a float.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a float variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(float& value);
	};
	
	class StringCfgVar : public CfgVar {
	private:
		//!	The variable's value.
		std::string m_value;

		StringCfgVar();

	public:
		//!	Constructor.
		/*!	This creates a string configuration variable and registers it
			into the variable table.
			\param name the variable's name.
			\param value the variable's initial value.
		*/
		StringCfgVar(const char* name, const std::string& value);

		//!	Destructor.
		virtual ~StringCfgVar();

		//!	String conversion operator.
		/*!	\return the variable's value.
		*/
		operator const std::string&() const { return m_value; }

		//!	C string conversion operator.
		/*!	\return the variable's value as a C string.
		*/
		operator const char*() const { return m_value.c_str(); }

		//!	Assignment operator.
		/*!	This operator invokes the set(string& value) method.
			\return this.
			\param src the value to be assigned.
		*/
		StringCfgVar& operator=(const std::string& src);

		//!	Get the variable's value as a string.
		/*!	\return true if the variable is a string variable, false otherwise.
			\param value used to store the variable's value.
		*/
		virtual bool get(std::string& value);

		//!	Set the variable's value as a string.
		/*!	This method will try to change the variable's value.
			If the proposed value is a correct value for the variable,
			the variable will be assigned a new value. Otherwise, the
			variable will act as described in its handler.
			\return true if the variable is a string variable, false otherwise.
			\param value The value to be assigned to the variable.
		*/
		virtual bool set(std::string& value);
	};
	
}

#endif
