/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: display.h,v 1.3 2003/08/28 13:48:22 rodzilla Exp $

    The Display class handles SDL and OpenGL initialisation as
    well as screenshots.
*/

#ifndef __SAM3D_DISPLAY_H__
#define __SAM3D_DISPLAY_H__

#include <SaM3d/config.h>
#include <SaM3d/config_variable.h>
#include <SaM3d/viewport.h>
#include <SDL/SDL.h>
#include <vector>

namespace SaM3d {

	//!	Video mode class.
	class VideoMode {
		//!	Video mode width.
		unsigned int m_w;

		//!	Video mode height.
		unsigned int m_h;
	
		VideoMode() { }

	public:
		//!	Constructor.
		VideoMode(const unsigned int w, const unsigned int h);

		//!	Copy constructor.
		VideoMode(const VideoMode& vm);

		//!	Return the width.
		unsigned int w() const { return m_w; }

		//!	Return the height.
		unsigned int h() const { return m_h; }

		//!	Equality operator.
		bool operator ==(const VideoMode& vm) const;

		//!	Comparison operator.
		bool operator <(const VideoMode& vm) const;

		//!	Assignment operator.
		VideoMode& operator =(const VideoMode& vm);
	};

	//!	Display class.
	/*!	This class handles the display, video mode changes and
		screenhots.
		This is a singleton class. Do not use the class itself,
		use the Display macro instead.
	*/
	class CDisplay {
		struct VideoModeHandler : public CfgVar::NumHandler {
			int m_max;

			VideoModeHandler(const int max);
			virtual bool operator ()(float& value) const;
			void set_max(const int max) { m_max = max; }
		};
	friend class VideoModeHandler;

		//!	A pointer to the single instance of this class.
		static CDisplay	*m_instance;

		//!	The SDL_Surface for the application.
		SDL_Surface*		m_surface;
		//!	The viewport for the whole application.
		Viewport		m_viewport;

		//!	The windowed mode width.
		unsigned int		m_window_width;
		//!	The windowed mode height.
		unsigned int		m_window_height;

		//!	An array containing the resolutions of all available fullscreen modes.
		std::vector<VideoMode>	m_video_modes;
		//!	The number of available fullscreen video modes.
		char			m_nb_video_modes;

		//! 	The default constructor.
		/*!	The constructor initializes a few members and invokes init_display().
			
		  	Note that it is protected because the CDisplay class is a singleton.
		*/
		CDisplay();

		//!	Initialize the display.
		/*!	This method does the video initialization.
			\return 0 everything is OK, 1 otherwise.
		*/
		char init_display();

		//!	Create the video mode list.
		/*!	This method is invoked by init_display().
		*/
		void build_video_mode_list();

		//!	Set the video mode.
		/*!	Change the video mode.
			\param mode the new video mode. 0 is windowed mode. mode > 0 is
			       video mode as listed using list_video_mode(). mode < 0
			       is incorrect (this will default to windowed).
		*/
		void set_video_mode(const char mode);

		VideoModeHandler m_video_mode_handler;

	public:
		//!	Obtain the instance to the CDisplay singleton.
		/*!	Don't bother ! Use the Display macro instead.
			\return a reference to the unique instance of CDisplay.
		*/
		static CDisplay& instance();

		//!	The current video mode.
		IntCfgVar video_mode;

		//!	Get the number of video modes.
		/*!	\return the number of available video modes.
		*/
		const char get_nb_video_modes() const;

		//!	Get info about a video mode.
		/*!	\param mode the video mode's id (an assertion will check that it is valid).
			\return the VideoMode structure corresponding to mode.
		*/
		const VideoMode& get_video_mode_info(const char mode) const;

		//!	Resize the window / change the video mode.
		/*!	Handle with care, it is meant for internal use.
		
			It is invoked internally and by Loop::start().
		  
		  	This method is used :
			- to change the video mode (with windowed set to false) ;
			- when the application's window is resized.
			It updates the underlying SDL_Surface and the main
			viewport of the application.
			\param w the width of the window/video mode.
			\param h the height of the window/video mode.
			\param windowed set it to false to set a new video
			                mode, to true when resizing the window.
			\todo protect this method dans make Loop a friend class (not
			      sure I will do this because it will make it impossible
			      to handle window resize events when not using Loop ;
			      time will tell !).
			      
		*/
		void resize_window(const unsigned int w, const unsigned int h,
				   const bool windowed);

		//!	Set the title of the application's window.
		/*!	\param text the text to be used as window title.
		*/
		void set_window_title(const char* text);

		//!	Return the display's viewport.
		/*!	\return the viewport associated to the display.
		*/
		const Viewport& viewport() const { return m_viewport; }

		//!	Take a screenshot.
		/*!	This method saves a screenshot into a BMP file.
			\param filename the name of the file in which the
					screenshot is saved.
			\return true if the screenshot succeeded, false otherwise.
		*/
		bool screenshot(const char* filename) const;
	};

}

//!	Use this macro to access the display.
#define Display (SaM3d::CDisplay::instance())

#endif
