/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2002 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: file.h,v 1.10 2003/09/03 11:13:45 rodzilla Exp $
*/

#ifndef __SAM3D_FILE_H__
#define __SAM3D_FILE_H__

#include <set>
#include <string>
#include <SDL/SDL.h>

namespace SaM3d {

	class BaseDir;
	class File;

	//!	Directory class.
	/*!	This is a class for virtual directories.
		A Dir consists of one or several directories or zip
		files.
	*/
	class Dir {
	friend class File;

		//!	The virtual directories.
		std::set<BaseDir*> m_dirs;

		//!	Open a virtual directory.
		/*!	\return the new virtual directory.
		 	\param dir The virtual directory in which the directory must be opened.
			\param path the (relative) path of the directory to be opened.
		*/
		static BaseDir* open_dir(BaseDir* dir, const char* path);

		Dir() { }
	public:
		//!	Constructor.
		/*!	The directory is opened from the current path.
			The pathlist parameter can consist of several paths using the ";"
			separator (in which case the Dir will consist of all opened
			virtual directories).
			\param pathlist the list of paths for the directory to be created.
		*/
		Dir(const char* pathlist);

		//!	Constructor.
		/*!	The directory is opened from the src directory (which means
			it will consists in one virtual directory per virtual directory
			in src and per path in pathlist).
			separator (in which case the Dir will consist of all opened
			virtual directories).
			\param src the source directory.
			\param pathlist the list of paths for the directory to be created.
		*/
		Dir(const Dir& src, const char* pathlist);

		//!	Copy constructor.
		/*!	\param src the directory to be copied.
		*/
		Dir(const Dir& src);

		//!	Destructor.
		~Dir();

		//!	Assignment operator.
		/*!	\param src the source for the assignment.
		*/
		Dir& operator =(const Dir& src);

		//!	Comparison operator.
		bool operator ==(const Dir& src) const;

		//!	Build a list of files in a path from a directory.
		/*!	\param file_list the file list to be filled.
			\param dir the (relative) path of the directory to be listed.
		*/
		void get_file_list(std::set<std::string>& file_list, const char* dir = NULL) const;
	};

	//!	File class.
	/*	This is a useful class to read files, basically a C++ wrapper for
		the fantastic SDL_RWops.
	*/
	class File {
		//!	A pointer to the corresponding SDL_RWops.
		SDL_RWops*	m_rwops;

		//!	A pointer to the corresponding buffer(to free it when destroying the file).
		void*		m_mem;

		//!	A boolean used to detect the end of file.
		bool		m_eof;

		File();
		File(const File&);

	public:
		//!	Constructor.
		/*!	Open a file given a directory, a filename and a
			list of accepted extensions.
			\param d the directory.
			\param s the filename.
			\param e the valid extensions.
		*/
		File(const Dir& d, const char* s, char** e = NULL);

		//!	Constructor.
		/*!	Open a file from the current directory, given a filename
			and a list of accepted extensions.
			\param s the filename.
			\param e the valid extensions.
		*/
		File(const char* s, char** e = NULL);

		//!	Destructor.	
		~File();

		//!	Seek in the file.
		/*!	This behaves as fseek().
			\todo Should look at how it impacts eof().
		*/
		int seek(int offset, int whence);

		//!	Get a character from the file.
		/*!	This behaves as fget().
		*/
		char get();

		//!	Read from a file.
		/*!	This behaves as fread().
		*/
		int read(void *ptr, int size, int maxnum);

		//!	Return true if the end of the file has been reached.
		/*!	This behaves as feof().
		*/
		bool eof();

		//!	Read an unsigned char.
		void read(unsigned char& x);

		//!	Read an unsigned short (little endian).
		void read_le(unsigned short& x);

		//!	Read an unsigned int (little endian).
		void read_le(unsigned int& x);

		//!	Read a string.
		void read(std::string& s, unsigned int l);

		//!	Get the SDL_RWops.
		SDL_RWops* RWops() { return m_rwops; }
	};

}

#endif
