/**************************************************************************

    This file is part of "The Science and Magick base library", a
    general library that provides miscellaneous features.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: log.h,v 1.3 2003/08/09 09:35:15 rodzilla Exp $
*/
#ifndef __SAM3D_LOG_H__
#define __SAM3D_LOG_H__

#include <fstream>
#include <stdarg.h>
#include <string>

namespace SaM3d {

	//!	A simple logging class.
	/*!	This class is useful to handle error/warning/miscellaneous
		message.
		It keeps the last 1024 m_max_lines in memory so they can be
		displayed in a console.
		\todo This class needs a way of writing the messages into a log file.
	*/
	class CLog {
		//!	The size of the message buffer.
		static const unsigned int m_max_lines = 1024;
		//!	The message buffer.
		std::string m_lines[m_max_lines];

		//!	The number of (non-empty) lines in the buffer.
		unsigned int m_nb_lines;
		//!	The line being written at the moment.
		unsigned int m_current_line;

		std::ofstream m_log_file;

		//!	Add a string to the log.
		/*!	This function chops the string (using the newline
			character as a separator) and invokes add_chars for
			each substring.
			\param s the string to be added to the message log.
		*/
		void add_string(char* s);

		//!	Add a string to the current line.
		/*!	\param s the string to be added to the current line.
		*/
		void add_chars(const char* s);

		//!	Increment the current line number.
		void endl();

		//!	The constructor for CLog.
		/*!	This class is a singleton : the constructor is private.
		*/
		CLog();

	public:
		//!	Create (if necessary) and return the unique instance of CLog.
		/*!	\return a reference to the unique instance of CLog.
		*/
		static CLog& instance(); 

		//!	Write an boolean to the message buffer.
		CLog& operator<<(const bool b);

		//!	Write an integer to the message buffer.
		CLog& operator<<(const int i);

		//!	Write a float to the message buffer.
		CLog& operator<<(const float f);

		//!	Write a double to the message buffer.
		CLog& operator<<(const double d);

		//!	Write a string to the message buffer.
		CLog& operator<<(const char* s);

		//!	Write a string to the message buffer.
		CLog& operator<<(const std::string& s);

		//!	Write a formatted string to the message buffer.
		void printf(const char* fmt, ...);

		//!	Write a formatted string to the message buffer.
		void vprintf(const char* fmt, va_list args);

		//!	Get the number of non-empty lines in the message buffer.
		const unsigned int get_nb_lines();

		//!	Get the contents of a line in the message buffer.
		const char* get_line(const unsigned int n);
	};
}

//! This macro returns the a reference to the unique CLog instance.
#define Log SaM3d::CLog::instance()

#endif
