/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: loop.h,v 1.4 2003/08/28 13:48:22 rodzilla Exp $

    A basic skeleton for loops (scenes ?).
*/

#ifndef __SAM3D_LOOP_H__
#define __SAM3D_LOOP_H__

#include <SaM3d/config.h>
#include <SaM3d/status.h>
#include <SDL/SDL.h>

namespace SaM3d {

	//!	OpenGL Loop class.
	/*!	This class is a base class for all OpenGL loops. It doesn't
	 	do anything except loop forever and wait for events.
		It processes keyboard, mouse and window manager events.
		Just define subclasses to create useful loops.
		To use a Loop subclass, create an instance and invoke the
		start() method.
	*/
	class Loop {
	friend class CDisplay;
	friend class Status;

		//!	Pointer to the current loop.
		static Loop* M_current_loop;
		//!	Current time.
		static Uint32 M_current_time;

		//!	Control variable for the loop.
		/*!	If true, the loop is about to exit.
		*/
		bool m_done;
		//!	Exit status for the loop.
		int m_exit_status;

	protected:
		//!	Get the status of a key.
		/*!	Return the status (pressed or not) of a key.
			\return 1 if the key is pressed, 0 otherwise.
			\param k the SDL key code of the key to be tested.
		*/
		static char key_state(const SDLKey k);

		//!	Get the status of a mouse button.
		/*!	Return the status (pressed or not) of a mouse button.
			\param b the number of the mouse button to be tested.
			\return 1 if the button is pressed, 0 otherwise.
		*/
		static char mouse_button_state(const Uint8 b);

		//!	Method invoked when a key is pressed.
		/*!	This method is invoked by the start() method
			whenever a key is pressed. Redefine it in a subclass
			to take keys into account.
			The return value is true if the key press has been processed,
			false othewise.
			\param keysym the SDL keycode of the key that has been
			       pressed.
			\return false (as this base loop class does nothing, this
			        function does nothing either).
		*/
		virtual bool handle_key_pressed(const SDL_keysym *keysym);

		//!	Method invoked when a key is released.
		/*!	This method is invoked by the start() method
			whenever a key is released. Redefine it in a subclass
			to take keys into account.
			The return value is true if the key release has been processed,
			false othewise.
			\param keysym the SDL keycode of the key that has been
			       released.
			\return false (as this base loop class does nothing, this
			        function does nothing either).
		*/
		virtual bool handle_key_released(const SDL_keysym *keysym);

		//!	Method invoked when a mouse button is pressed.
		/*!	This method is invoked by the start() method
			whenever a mouse button is pressed. Redefine it in a subclass
			to take mouse buttons into account.
			The return value is true if the mouse button press has been
			processed, false othewise.
			\param button the number of the mouse button that has been
			       pressed.
			\return false (as this base loop class does nothing, this
			        function does nothing either).
		*/
		virtual bool handle_mouse_button_pressed(const Uint8 button);

		//!	Method invoked when a key is released.
		/*!	This method is invoked by the start() method
			whenever a mouse button is released. Redefine it in a subclass
			to take mouse buttons into account.
			The return value is true if the mouse button release has been
			processed, false othewise.
			\param button the number of the mouse button that has been
			       released.
			\return false (as this base loop class does nothing, this
			        function does nothing either).
		*/
		virtual bool handle_mouse_button_released(const Uint8 button);
	
		//!	Method invoked when the mouse is moved.
		/*!	This method is invoked by the start() method
			whenever the mouse is moved. Redefine it in a subclass
			to take mouse motion into account.
			The return value is true if the mouse motion has been
			processed, false othewise.
			\param x the horizontal motion of the mouse.
			\param y the vertical motion of the mouse.
			\return false (as this base loop class does nothing, this
			        function does nothing either).
			\todo x and y could be floats and calculated from real motion
			      and mouse sensitivity.
		*/
		virtual bool handle_mouse_motion(const int x, const int y);
		
		//!	Method invoked when the window is resized.
		/*!	This method is invoked by the start() method
			whenever the window is resized. Redefine it in a subclass
			to take window resize into account.
			The return value is true if the window resize has been
			processed, false othewise.
			\param w the new width of the window.
			\param h the new height of the window.
			\return false (as this base loop class does nothing, this
			        function does nothing either).
			\todo x and y could be floats and calculated from real motion
			      and mouse sensitivity.
		*/
		virtual bool handle_window_resize(const int w, const int h);

		//!	Method invoked whenever the loop starts.
		/*!	This method is invoked first when start is invoked,
		        so this is the right place for all kinds of OpenGL
			initializations.
		*/
		virtual void init();

		//!	Drawing method invoked once per frame.
		/*!	This method is where all OpenGL rendering must be implemented.
		*/
		virtual void draw();

		//!	Motion method invoked once per frame.
		/*!	This method is where all motion must be implemented.
			\param dt time elapsed since the previous frame (in seconds).
		*/
		virtual void move(float dt);

		//!	Exit the loop.
		/*!	This method simply ends the loop. The loop will exit after
			the the end of the current iteration.
		*/
		void exit(int exit_status = 0);

	public:
		//!	A virtual destructor.
		/*!	Remember when creating a subclass of Loop, the destructor
			should be virtual (not REALLY necessary when using Loop
			the simple way, but who knows ?).
		*/
		virtual ~Loop() {}

		//!	The loop itself.
		/*!	This method invokes init() and starts a loop in which :
			- it processes events and invokes the corresponding methods
			  (see handle_key_pressed(), handle_key_released(),
                	  handle_mouse_button_pressed(), handle_mouse_button_released(),
                	  handle_mouse_motion() and handle_window_resize()).
			- it invokes the draw() method.
			- calculates the time elapsed since the previous frame and
			  invokes the move() method.
			The loop will exit whenever the exit() function is invoked.
		*/
		int start();

		//!	Get the current time.
		static Uint32 current_time() { return M_current_time; }
	};

}

#endif
