/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: math.h,v 1.2 2003/08/03 15:40:56 rodzilla Exp $

    Math-related classes. Vectors, Matrices, Quaternions, Frustums.
*/

#ifndef __SAM3D_MATH_H__
#define __SAM3D_MATH_H__

#include <SaM3d/config.h>
#include <math.h>
#include <iostream>
#include <GL/gl.h>

#ifndef M_PI
#define M_PI 3.1415926535898
#endif

namespace SaM3d {

	//!	Vector class.
	/*!	This class implements a basic 2-components vector.
	*/
	class Vector2 {
	public:
		//!
		float x;
		//!
		float y;

		//!	Default constructor : create null vector.
		Vector2() : x(0.0), y(0.0) { }

		//!	Basic constructor : create a vector from its components.
		/*!	\param _x x component value.
			\param _y y component value.
		*/
		Vector2(const float _x, const float _y) : x(_x), y(_y) { }

		//!	Copy constructor.
		/*!	\param v the vector to be copied.
		*/
		Vector2(const Vector2& v) : x(v.x), y(v.y) { }

		//!	Assignment operator.
		/*!	\param v the value to be assigned.
			\return a reference to this.
		*/
		Vector2& operator =(const Vector2& v) {
			x = v.x;
			y = v.y;
			return *this;
		}

		//!	Comparison operator.
		/*!	\param v the vector to be compared to this.
			\return true if v equals this, false otherwise. 
		 */
		bool operator ==(const Vector2& v) const {
			return (x == v.x) && (y == v.y);
		}

		//!	Comparison operator.
		/*!	\param v the vector to be compared to this.
			\return false if v equals this, true otherwise. 
		 */
		bool operator !=(const Vector2& v) const {
			return (x != v.x) || (y != v.y);
		}

		//!	Vector addition.
		/*!	This method calculates the sum of 2 vectors.
			\param v the second operand of the addition.
			\return the sum of this and v.
		*/
		Vector2 operator +(const Vector2& v) const {
			Vector2 rv(x + v.x, y + v.y);

			return rv;
		}

		//!	Vector substraction.
		/*!	This method the result of the substraction of 2 vectors.
			\param v the second operand of the substraction.
			\return this minus v.
		*/
		Vector2 operator -(const Vector2& v) const {
			Vector2 rv(x - v.x, y - v.y);

			return rv;
		}
		//!	Dot product.
		/*!	This method calculates the dot product of 2 vectors.
			\param v the second operand of the dot product.
			\return the dot product of this and v.
		*/
		float operator *(const Vector2& v) const {
			return x*v.x+y*v.y;
		}

		//!	Unary minus operator.
		/*!	\return the opposite of this.
		*/
		Vector2 operator -() const {
			Vector2 rv(-x, -y);

			return rv;
		}

		//!	Add and assign operator.
		/*!	This method adds a vector to this and returns a reference
			to this.
			\param v the second operand of the operator.
			\return a reference to this.
		*/
		Vector2& operator +=(const Vector2& v) {
			x += v.x;
			y += v.y;
			return *this;
		}

		//!	Substract and assign operator.
		/*!	This method substract a vector from this and returns a
			reference to this.
			\param v the second operand of the operator.
			\return a reference to this.
		*/
		Vector2& operator -=(const Vector2& v) {
			x -= v.x;
			y -= v.y;
			return *this;
		}

		//!	Calculate the squared length of a vector.
		/*!	\return the squared length of this.
		*/
		float length2() const {
			return x*x+y*y;
		}

		//!	Calculate the length of a vector.
		/*!	\return the length of this.
		*/
		float length() const {
			return sqrt(length2());
		}

		//!	Calculate a normalized copy of a vector.
		/*!	\return a normalized copy of this.
		*/
		Vector2 norm() const {
			float l2 = length2();

			if (l2 != 0.0 && l2 != 1.0) {
				float il=1/sqrt(l2);
				Vector2 rv(x*il, y*il);

				return rv;
			}

			return *this;
		}
	};

	//!	Multiplies a float and vector.
	/*!	\param f a float.
		\param v a Vector2.
		\return the product of f and v.
	*/
	inline Vector2 operator *(const float f, const Vector2& v) {
		Vector2 rv(f*v.x, f*v.y);
		return rv;
	}

	//!	Vector class.
	/*!	This class implements a basic 3-components vector.
	*/
	class Vector3 {
	public:
		//!
		float x;
		//!
		float y;
		//!
		float z;

		//!	Default constructor : create null vector.
		Vector3() : x(0.0), y(0.0), z(0.0) { }

		//!	Basic constructor : create a vector from its components.
		/*!	\param _x x component value.
			\param _y y component value.
			\param _z z component value.
		*/
		Vector3(const float _x, const float _y, const float _z) : x(_x), y(_y), z(_z) { }

		//!	Copy constructor.
		/*!	\param v the vector to be copied.
		*/
		Vector3(const Vector3& v) : x(v.x), y(v.y), z(v.z) { }

		//!	Assignment operator.
		/*!	\param v the value to be assigned.
			\return a reference to this.
		*/
		Vector3& operator =(const Vector3& v) {
			x = v.x;
			y = v.y;
			z = v.z;
			return *this;
		}

		//!	Comparison operator.
		/*!	\param v the vector to be compared to this.
			\return true if v equals this, false otherwise. 
		 */
		bool operator ==(const Vector3& v) const {
			return (x == v.x) && (y == v.y) && (z == v.z);
		}

		//!	Comparison operator.
		/*!	\param v the vector to be compared to this.
			\return false if v equals this, true otherwise. 
		 */
		bool operator !=(const Vector3& v) const {
			return (x != v.x) || (y != v.y) || (z != v.z);
		}

		//!	Vector addition.
		/*!	This method calculates the sum of 2 vectors.
			\param v the second operand of the addition.
			\return the sum of this and v.
		*/
		Vector3 operator +(const Vector3& v) const {
			Vector3 rv(x + v.x, y + v.y, z + v.z);

			return rv;
		}

		//!	Vector substraction.
		/*!	This method the result of the substraction of 2 vectors.
			\param v the second operand of the substraction.
			\return this minus v.
		*/
		Vector3 operator -(const Vector3& v) const {
			Vector3 rv(x - v.x, y - v.y, z - v.z);

			return rv;
		}
		//!	Dot product.
		/*!	This method calculates the dot product of 2 vectors.
			\param v the second operand of the dot product.
			\return the dot product of this and v.
		*/
		float operator *(const Vector3& v) const {
			return x*v.x+y*v.y+z*v.z;
		}

		//!	Cross product.
		/*!	This method calculates the cross product of 2 vectors.
			\param v the second operand of the cross product.
			\return the cross product of this and v.
		*/
		Vector3 operator ^(const Vector3& v) const {
			Vector3 rv(y*v.z-z*v.y,
				   z*v.x-x*v.z,
				   x*v.y-y*v.x);

			return rv;
		}

		//!	Unary minus operator.
		/*!	\return the opposite of this.
		*/
		Vector3 operator -() const {
			Vector3 rv(-x, -y, -z);

			return rv;
		}

		//!	Add and assign operator.
		/*!	This method adds a vector to this and returns a reference
			to this.
			\param v the second operand of the operator.
			\return a reference to this.
		*/
		Vector3& operator +=(const Vector3& v) {
			x += v.x;
			y += v.y;
			z += v.z;
			return *this;
		}

		//!	Substract and assign operator.
		/*!	This method substract a vector from this and returns a
			reference to this.
			\param v the second operand of the operator.
			\return a reference to this.
		*/
		Vector3& operator -=(const Vector3& v) {
			x -= v.x;
			y -= v.y;
			z -= v.z;
			return *this;
		}

		//!	Calculate the squared length of a vector.
		/*!	\return the squared length of this.
		*/
		float length2() const {
			return x*x+y*y+z*z;
		}

		//!	Calculate the length of a vector.
		/*!	\return the length of this.
		*/
		float length() const {
			return sqrt(length2());
		}

		//!	Calculate a normalized copy of a vector.
		/*!	\return a normalized copy of this.
		*/
		Vector3 norm() const {
			float l2 = length2();

			if (l2 != 0.0 && l2 != 1.0) {
				float il=1/sqrt(l2);
				Vector3 rv(x*il, y*il, z*il);

				return rv;
			}

			return *this;
		}

		//!	Invoke glVertex with this's components.
		void glVertex() const;

		//!	Invoke glNormal with this's components.
		void glNormal() const;

		//!	Invoke glTranslate with this's components.
		/*	Useful for transforming the model view matrix when
			this is the position of a displayed object.
		*/
		void glTranslate() const;

		//!	Invoke glTranslate with -this's components.
		/*	Useful for transforming the model view matrix when
			this is the position of a camera.
		*/
		void inv_glTranslate() const;
	};

	//!	Multiplies a float and vector.
	/*!	\param f a float.
		\param v a Vector3.
		\return the product of f and v.
	*/
	inline Vector3 operator *(const float f, const Vector3& v) {
		Vector3 rv(f*v.x, f*v.y, f*v.z);
		return rv;
	}

	//!	Plane class.
	/*!	This is useful when (for example) partitioning a scene.
	*/
	class Plane {
	public:
		//!	x coordinate of the normal vector.
		float a;
		//!	y coordinate of the normal vector.
		float b;
		//!	z coordinate of the normal vector.
		float c;
		//!	distance to the origin (provided (x, y, z) is normalized).
		float d;

		Plane();
		Plane(const Plane& p);

		Plane& operator =(const Plane& p);

		//!	Check if a point is in front of the plane.
		bool front(const Vector3& v) const;
	};

	class Quaternion;

	//!	3x3 Matrix class.
	/*!	This kind of matrices is useful for representing rotations, scaling...
	*/
	class Matrix3x3 {
	friend class Quaternion;

		GLfloat m_data[9];

		//!	Default constructor.
		/*!	It cannot be invoked. Initialize your matrices !
			\todo OK, it could be public and create an identity matrix.
		*/
		Matrix3x3() { }

	public:
		//!	Copy constructor.
		/*!	\param m the matrix to be copied.
		*/
		Matrix3x3(const Matrix3x3& m);

		//!	Constructor.
		/*!	This constructor creates a Matrix3x3 from a quaternion.
			The obtained matrix represents the same rotation as the
			quaternion.
			\param q the quaternion.
		*/
		Matrix3x3(const Quaternion& q);

		//!	Assignment operator.
		/*!	\param m the value to be assigned.
			\return a reference to this.
		*/
		Matrix3x3& operator =(const Matrix3x3& m);

		//!	Matrix-vector product.
		/*!	This method calculates the product of a matrix and a vector.
			This is useful to apply a rotation to a vector.
			\param v the vector to be rotated.
			\return the rotated vector.
		*/
		Vector3 operator *(const Vector3& v) const;
	};

	//!	Quaternion class.
	/*!	Quaternions are used to repesent rotations (vector rotations, not
		Vertex rotations).
		\todo There are quite a few useful things missing here (including SLERP).
		      I'll add them whenever I need them.
	*/
	class Quaternion {
	friend class Matrix3x3;

		float m_w, m_x, m_y, m_z;
	public:

		//!	Default constructor.
		/*!	It cannot be invoked. Initialize your quaternions !
		*/
		Quaternion();

		//!	Base constructor.
		/*!	Creates a quaternion starting from a vector (axis of the rotation)
			and an angle.
			\param v axis of the rotation.
			\param a angle of the rotation.
		*/
		Quaternion(const Vector3& v, const float a);

		//!	Copy constructor.
		/*!	\param q the quaternion to be copied.
		*/
		Quaternion(const Quaternion& q);

		//!	Assignment operator.
		/*!	\param q the value to be assigned.
			\return a reference to this.
		*/
		Quaternion& operator =(const Quaternion& q);

		//!	Multiply and assign operator.
		/*!	This method multiplies this with another quatenion.
			It is useful for composing rotations.
			\param q the second operand of the operator.
			\return a reference to this.
		*/
		Quaternion& operator *=(const Quaternion& q);

		//!	Inverts a quaternion.
		/*!	\return the opposite of this.
		*/
		Quaternion operator -() const;

		//*	Product operator.
		/*!	This method calculates the product of this with another quatenion.
			It is useful for composing rotations.
			\param q the second operand of the operator.
			\return The presult of the product.
		*/
		Quaternion operator *(const Quaternion& q) const;

		//!	Invoke glRotate with this.
		/*	Useful for transforming the model view matrix when
			this is the position of a displayed object.
		*/
		void glRotate() const;

		//!	Invoke glRotate with -this.
		/*	Useful for transforming the model view matrix when
			this is the position of a camera.
		*/
		void inv_glRotate() const;
	};

	//!	A simple frustum class.
	/*!	This class can extract the current frustum to test the
		visibility of a point, a sphere or a box.
	*/
	class Frustum {
		//!	The data that represents the frustum (the equations of the 6 planes).
		float m_data[6][4];

	public:
		//!	Update the frustum so that it matches the current view.
		void update();

		//!	Test is a point is in the frustum.
		bool test_point(const Vector3&) const;

		//!	Test is a sphere is in or intersects the frustum.
		bool test_sphere(const Vector3&, const float) const;

		//!	Test is a box is in or intersects the frustum.
		bool test_box(const Vector3& mins, const Vector3& maxs) const;
	};

}

#endif

