/**************************************************************************

    This file is part of "The Science and Magick base library", a
    general library that provides miscellaneous features.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: memory.h,v 1.3 2003/08/28 13:48:22 rodzilla Exp $
*/
#ifndef __SAM3D_MEMORY_H__
#define __SAM3D_MEMORY_H__

#include <stdlib.h>
#include <assert.h>

namespace SaM3d {

	//!	A basic ref counted class.
	class RefCounted {
	protected:
		//!	The number of references.
		unsigned m_references ;

		//!	The function invoked whenever the last reference is removed.
		virtual void free() { delete this; }

	public:
		//!	Constructor.
		/*!	Initializes the number of references to 0.
		*/
		RefCounted() : m_references(0) { }

		//!	Destructor.
		virtual ~RefCounted() { }

		//!	Add a reference.
		void add_ref() { ++m_references; }
		//!	Removes a reference.
		void rem_ref() { if (--m_references == 0) free(); }
	};

	//!	A smart pointer class that handles reference counts.
	template <class T>
	class SmartPtr {
	protected:
		//!	The pointer itself.
		T* m_ptr;

	public:
		//!	Constructor.
		/*!	Adds a reference to the pointed object if any
			\param t the initial value of the pointer.
		*/
		SmartPtr(T* t = NULL) : m_ptr(t) {
			if (m_ptr) m_ptr->add_ref();
		}

		//!	Copy constructor.
		SmartPtr(const SmartPtr<T>& p) : m_ptr(p.m_ptr) {
			if (m_ptr) m_ptr->add_ref();
		}

		//!	Destructor.
		virtual ~SmartPtr() {
			if (m_ptr) m_ptr->rem_ref();
		}

		//!	Assignment operator.
		/*!	\param t the new value for the pointer.
		*/
		SmartPtr<T>& operator=(T* t) {
			if (t != m_ptr) {
				if (m_ptr) m_ptr->rem_ref();
				m_ptr = t;
				if (m_ptr) m_ptr->add_ref();
			}
			return *this;
		}

		//!	Assignment operator.
		/*!	\param p the new value for the pointer.
		*/
		SmartPtr<T>& operator=(const SmartPtr<T>& p) {
			if (p.m_ptr != m_ptr) {
				if (m_ptr) m_ptr->rem_ref();
				m_ptr = p.m_ptr;
				if (m_ptr) m_ptr->add_ref();
			}
			return *this;
		}

		//!	Cast operator.
		/*!	Returns a "real" pointer.
		*/
		operator T*() const { return m_ptr; }

		//!	Dereferencing operator.
		T* operator ->() const { assert(m_ptr != NULL); return m_ptr; }
	};

}

#endif
