/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: texture.h,v 1.10 2003/08/28 13:48:22 rodzilla Exp $

    Classes for textures.
*/

#ifndef __SAM3D_TEXTURE_H__
#define __SAM3D_TEXTURE_H__

#include <SaM3d/config.h>
#include <SaM3d/memory.h>
#include <SaM3d/opengl.h>
#include <set>
#include <string>

namespace SaM3d {

	class BaseTexture;

	//!	Smart pointer to a BaseTexture.
	/*!	This is the class to be used to manipulate textures.
	*/
	class Texture : public SmartPtr<BaseTexture> {
	friend class BaseTexture;
	friend void SaM3d::OpenGL::init();
#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
	friend void OpenGL::clear_context();
#endif

	public:
		class Manager;

	/* Multi-texture features */ 
	private:
		//!	True is the OpenGL has support for multitexture.
		static bool M_has_multitexture;
		//!	Number of available texture units.
		static GLint M_nb_tex_units;
		//!	Current texture unit.
		static unsigned char M_texture_unit;
		//!	Current texture in each texture unit.
		static BaseTexture* M_current_textures[32];

	public:
		//!	Get the number of available texture units.
		/*!	\return the number of available texture units.
		*/
		static GLint nb_texture_units() { return M_nb_tex_units; }

		//!	Change the active texture unit.
		/*!	This function calls the glActiveTexture and
			glClientActiveTexture to change the active texture
			unit if necessary.
			\param u the index of the texture unit to be activated.
		*/
		static void set_texture_unit(const unsigned char u);

	/* Image formats support */ 
	private:
		//!	True if the OpenGL driver has support for the BGRA format.
		static bool M_bgra_support;

	public:
		//!	Return true if the OpenGL driver supports the BGRA format.
		static bool has_bgra_support() { return M_bgra_support; }

	/* Sphere mapping */
	public:
		//!	Enables/disables sphere mapping.
		static void tc_gen(const GLenum mode);

#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
	protected:
		//!	Release all textures.
		/*!	This method must be called under Windows just
			before changing resolution or resizing the window
			to make sure that all textures will be re-created.
		*/
		static void release_all();
#endif

	public:
		//*	Return the "none" texture.
		/*!	This method returns a texture that corresponds to
			"no texture". Binding this texture is the same as
			disabling texturing.
		*/
		static const Texture& none();

	/* Class contents */
	public:
		//!	Default constructor.
		Texture();

		//!	Constructor.
		/*!	This constructor creates a texture from a pointer to
			a BaseTexture. It will register the BaseTexture into the
			Texture::Manager or destroy it is a doppleganger is already
			registered. This mechanism makes sure all textures are unique.
			This also adds a reference to the pointed BaseTexture.
			\param t the BaseTexture to be registered into the Texture::Manager.
		*/
		Texture(BaseTexture* t);

		//!	Copy constructor.
		/*!	Creates a copy of a Texture and adds a reference to its pointed
			BaseTexture.
		*/
		Texture(const Texture& t);

		//!	Destructor.
		/*!	Removes a reference to the pointed BaseTexture.
			\note Whenever a BaseTexture gets out of reference, it will
			      unregister from the Texture::Manager and delete itself.
		*/
		virtual ~Texture();

		//!	Assignment operator.
		/*!	This operator changes the pointed BaseTexture, removes
			a reference from the previously pointed BaseTexture and
			adds a reference to the new one.
			\param t the new BaseTexture to be pointed to.
			\return a reference to this.
		*/
		Texture& operator=(BaseTexture* t);

		//!	Bind the texture.
		/*!	This function replaces glEnable(GL_TEXTURE_XD),
			glDisable(GL_TEXTURE_XD) and glBindTexture(...).
			If the pointed texture is the none() texture, this
			method disables texturing. Otherwise, it enables
			texturing and creates/updates then binds the
			corresponding BaseTexture.
		*/
		inline void bind() const;

		//!	Get the texture's ID.
		inline const GLuint id() const;
	};

	//!	Class ids for texture classes (useful to compare textures).
	enum class_id_t {
		id_base_texture, id_file_texture, id_lightmap
	};

	//!	The base class for all kinds of textures.
	/*!	\todo This class should only take care of texture-related
		things. Two other classes must be added :
		- ManagedTexture : textures handled by the texture manager,
		  i.e. textures that require unicity.
		- SpecialTextures : textures that are used for part of the
		  rendering process (i.e. render-to-texture for example).
		  Those do not need the unicity/ref-counting things.
	*/
	class BaseTexture : public RefCounted {
	friend class SaM3d::Texture::Manager;

	public:
		//!	Return the class id of the BaseTexture.
		virtual unsigned is_a() const { return id_base_texture; }

		//!	Compares two BaseTextures.
		virtual bool equal(const BaseTexture* t) const;

		//!	Return the hash value of the BaseTexture.
		/*!	This is required by the Texture::Manager.
		*/
		virtual unsigned char hash() const;

	protected:
		//!	Possible values for the BaseTexture's status.
		enum status_t { uninitialized, ok, no_texture, invalid };

		//!	OpenGL ID for the texture.
		GLuint		m_id;

		//!	Status of the BaseTexture.
		/*!	This tells if the texture needs to be created/updated.
		*/
		status_t	m_status;

		//!	Create or update the texture.
		/*!	This method is defined in all child classes.
		*/
		virtual void create();

		//!	Release the texture.
		virtual void release();

	public:
		//!	Constructor.
		BaseTexture();

		//!	Destructor.
		virtual ~BaseTexture();

	private:
		//!	Unregister the texture and delete it.
		virtual void free();

	public:
		//!	Bind the texture.
		void bind();

		//!	Get the texture's ID.
		const GLuint id() const { return m_id; }
	
	};

	void
	Texture::
	bind() const {
		m_ptr->bind();
	}

	const GLuint
	Texture::
	id() const {
		return m_ptr->id();
	}

}

#endif
