/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: camera.cpp,v 1.1.1.1 2003/07/25 19:28:17 rodzilla Exp $

    The Camera class defines a basic camera.
*/

#include <unistd.h>
#include <SaM3d/camera.h>
#include <SaM3d/opengl.h>
#include <SaM3d/viewport.h>

namespace SaM3d {

	// The current camera is initialized to NULL.
	Camera* Camera::m_current = NULL;

	// The default construtor initializes the forward, up and right vectors.
	Camera::
	Camera() : m_near(1),
		   m_far(1000),
		   m_angle(45),
		   m_forward(0, 0, -1),
		   m_up(0, 1, 0),
		   m_right(1, 0, 0) {
	}

	// Moving the camera means adding a given vector to its position.
	void
	Camera::
	move(const Vector3& v) {
		m_position += v;
	}

	// Rotate the camera :
	void
	Camera::
	rotate(const Quaternion& q) {
		// apply the rotation to the orientation.
		m_orientation *= q;

		// recalculate the forward, up and right vectors.
		{
			Matrix3x3 m(-m_orientation);
			
			m_forward = m*Vector3(0, 0, -1);
			m_up = m*Vector3(0, 1, 0);
			m_right = m*Vector3(1, 0, 0);
		}
	}

	// Set the camera's position.
	void
	Camera::
	set_position(const Vector3& v) {
		m_position = v;
	}

	// Set the camera's field of view.
	void
	Camera::
	set_fov(const float n, const float f) {
		m_near = n;
		m_far = f;
	}

	// Set the camera's view angle.
	void
	Camera::
	set_angle(const float angle) {
		m_angle = angle;
	}

	// activate the camera.
	void
	Camera::
	activate() {
		float ratio = (float)Viewport::current().w()/(float)Viewport::current().h();

		// Update the projection matrix.
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(m_angle, ratio, m_near, m_far);

		// Update the model view matrix.
		// Here, the translation and rotation are inverted :
		// we're moving a camera, not an object.
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();		// Set it to identity.
		m_orientation.inv_glRotate();	// Apply the inverted "m_orientation" rotation 
		m_position.inv_glTranslate();   // Apply the inverted "m_position" translation
		m_current = this;
	}

}

