/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: config_variable.cpp,v 1.5 2003/08/24 15:45:00 rodzilla Exp $

    The Mesh class handle meshes.
*/

#include <map>
#include <SaM3d/log.h>
#include <SaM3d/config_variable.h>

namespace SaM3d {

	using namespace std;

	namespace {
		
		struct ltstr {
			bool operator()(const char* s1, const char* s2) const {
				return strcmp(s1, s2) < 0;
			}
		};

		map<const char*, CfgVar*, ltstr>&
		config_variables() {
			static map<const char*, CfgVar*, ltstr> vars;
			return vars;
		}
	}

	CfgVar::
	CfgVar(const char* name) {
		m_name = name;
		assert(config_variables().find(m_name) == config_variables().end());
		config_variables()[m_name] = this;
	}

	CfgVar::
	~CfgVar() {
	}

	CfgVar*
	CfgVar::
	get(const char* n) {
		map<const char*, CfgVar*, ltstr>::iterator i = config_variables().find(n);

		if (i == config_variables().end()) {
			return NULL;
		} else {
			return (*i).second;
		}
	}






	BoolCfgVar::
	BoolCfgVar(const char* name,
		   const bool value,
		   bool (*f)(bool&)) : CfgVar(name) {
		m_handler = f;
		*this = value;
	}

	BoolCfgVar::
	~BoolCfgVar() {
	}

	BoolCfgVar&
	BoolCfgVar::
	operator=(const bool src) {
		bool tmp = src;

		set(tmp);
		return *this;
	}

	bool
	BoolCfgVar::
	get(bool& src) {
		src = m_value;
		return true;
	}

	bool
	BoolCfgVar::
	set(bool& src) {
		if (   (m_handler != NULL) 
		    && (!m_handler(src))) {
			Log << "Invalid value for configuration variable " << m_name << ".\n";
			return false;
		}
		m_value = src;
		return true;
	}





	namespace {
		CfgVar::NumHandler&
		default_num_handler() {
			static CfgVar::NumHandler h;

			return h;
		}
	}

	IntCfgVar::
	IntCfgVar(const char* name,
		  const int value) : CfgVar(name),
				     m_handler(default_num_handler()) {
		*this = value;
	}

	IntCfgVar::
	IntCfgVar(const char* name,
		  const int value,
		  CfgVar::NumHandler& h) : CfgVar(name),
					   m_handler(h) {
		*this = value;
	}

	IntCfgVar::
	~IntCfgVar() {
	}

	IntCfgVar&
	IntCfgVar::
	operator=(const int src) {
		float tmp = src;

		set(tmp);
		return *this;
	}

	bool
	IntCfgVar::
	get(float& src) {
		src = (float)m_value;
		return true;
	}

	bool
	IntCfgVar::
	set(float& src) {
		if (!m_handler(src)) {
			Log << "Invalid value for configuration variable " << m_name << ".\n";
			return false;
		}
		m_value = (int)src;
		return true;
	}





	FloatCfgVar::
	FloatCfgVar(const char* name,
		  const float value) : CfgVar(name),
				     m_handler(default_num_handler()) {
		*this = value;
	}

	FloatCfgVar::
	FloatCfgVar(const char* name,
		  const float value,
		  CfgVar::NumHandler& h) : CfgVar(name),
					   m_handler(h) {
		*this = value;
	}

	FloatCfgVar::
	~FloatCfgVar() {
	}

	FloatCfgVar&
	FloatCfgVar::
	operator=(const float src) {
		float tmp = src;

		set(tmp);
		return *this;
	}

	bool
	FloatCfgVar::
	get(float& src) {
		src = (float)m_value;
		return true;
	}

	bool
	FloatCfgVar::
	set(float& src) {
		if (!m_handler(src)) {
			Log << "Invalid value for configuration variable " << m_name << ".\n";
			return false;
		}
		m_value = (float)src;
		return true;
	}




	StringCfgVar::
	StringCfgVar(const char* name,
		     const string& value) : CfgVar(name) {
		*this = value;
	}

	StringCfgVar::
	~StringCfgVar() {
	}

	StringCfgVar&
	StringCfgVar::
	operator=(const string& src) {
		string tmp = src;

		set(tmp);
		return *this;
	}

	bool
	StringCfgVar::
	get(string& src) {
		src = m_value;
		return true;
	}

	bool
	StringCfgVar::
	set(string& src) {
		m_value = src;
		return true;
	}

}

