/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: console.cpp,v 1.4 2003/08/09 10:07:51 rodzilla Exp $

    The CConsole class is a basic console that logs everything.
*/

#include <unistd.h>
#include <stdarg.h>
#include <SaM3d/log.h>
#include <SaM3d/config_variable.h>
#include <SaM3d/console.h>
#include <SaM3d/font.h>
#include <SaM3d/opengl.h>
#include <SaM3d/viewport.h>

namespace SaM3d {

	using namespace std;

	CConsole* CConsole::m_instance = NULL;

	CConsole::
	CConsole() {
		m_position = -1;
		m_mode = 0;
		m_interpreter = NULL;

		SDL_EnableUNICODE(1);
	}

	CConsole::
	~CConsole() {
	}

	void
	CConsole::
	set_interpreter(Interpreter* interpreter) {
		m_interpreter = interpreter;
	}

	namespace {

		IntCfgVar font_size("console_font_size", 12);

	}

	void
	CConsole::
	draw() {
		if (m_mode == 0) return;

		unsigned int i;
		const Viewport& vp = Viewport::current();
		unsigned int size_pixel_w;
		unsigned int size_pixel_h;
		unsigned int position;
		unsigned int text_position_y;
		unsigned int text_width;

		size_pixel_w = vp.w()-10;
		size_pixel_h = vp.h()-10;
		text_width = (size_pixel_w-10)/font_size;
		m_text_height = (size_pixel_h-10)/font_size;
		if (m_mode == 2) {
			m_text_height >>= 2;
		}
		size_pixel_h = (m_text_height+1)*font_size+2;
		m_text_height -= 2;

		if (m_position == -1) {
			if (Log.get_nb_lines() > m_text_height) {
				position = Log.get_nb_lines()-m_text_height-1;
			} else {
				position = 0;
			}
		} else {
			position = m_position;
		}

		glPushAttrib(GL_ALL_ATTRIB_BITS);

		vp.begin_view_ortho();
		glLoadIdentity();
		glDepthMask(GL_FALSE);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.2, 0.2, 0.4, 0.95);
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_TRIANGLE_STRIP);
		glVertex2i(5, vp.h()-5);
		glVertex2i(5, vp.h()-size_pixel_h+2*font_size);
		glVertex2i(vp.w()-5, vp.h()-5);
		glVertex2i(vp.w()-5, vp.h()-size_pixel_h+2*font_size);
		glEnd();
		glBegin(GL_TRIANGLE_STRIP);
		glVertex2i(5, vp.h()-size_pixel_h+font_size+5);
		glVertex2i(5, vp.h()-size_pixel_h);
		glVertex2i(vp.w()-5, vp.h()-size_pixel_h+font_size+5);
		glVertex2i(vp.w()-5, vp.h()-size_pixel_h);
		glEnd();
		glColor4f(1, 1, 1, 1);
		for (i = position, text_position_y = vp.h()-10;
		     i < position+m_text_height; ++i) {
			m_font.print(10, text_position_y, font_size, Log.get_line(i));
			text_position_y -= font_size;
		}
		text_position_y -= font_size;
		m_font.print(10, text_position_y, font_size, ("> " + m_command_line).c_str());
		vp.end_view_ortho();

		glPopAttrib();
	}

	bool
	CConsole::
	handle_key_pressed(const SDL_keysym *keysym) {
		if (keysym->sym == SDLK_F11) {
			if (++m_mode == 3) m_mode = 0;
			return true;
		} else if (m_mode != 0) {
			switch (keysym->sym) {
				case SDLK_PAGEUP:
					if (m_position == -1) {
						m_position = Log.get_nb_lines()-3*(m_text_height>>1);
					} else {
						m_position -= m_text_height>>1;
					}
					if (m_position < 0) m_position = 0;
					return true;

				case SDLK_PAGEDOWN:
					if (m_position != -1) {
						m_position += m_text_height>>1;
						if (m_position+m_text_height >= Log.get_nb_lines()) {
							m_position = -1;
						}
					}
					return true;

				case SDLK_BACKSPACE: {
					int l = m_command_line.length();
					if (l > 0) {
						m_command_line.erase(l-1);
					}
				}	return true;

				case SDLK_RETURN: {
					string command_line(m_command_line);

					m_command_line = "";
					Log << "\n> " << command_line << "\n";
					if (m_interpreter) {
						m_interpreter->do_string(command_line);
					} else {
						Log << "No interpreter loaded.\n";
					}
					return true;
				}

				default: {
					unsigned char key = keysym->unicode;

					if (   (key >= 32)
					    && (key <= 150)) {
						m_command_line += key;
						return true;
					}
					break;
				}
			}
		}
		return false;
	}

	bool
	CConsole::
	is_active() const {
		return (m_mode != 0);
	}

	CConsole&
	CConsole::
	instance() {
		if (m_instance == NULL) {
			m_instance = new CConsole();
		}
		return *m_instance;
	}

}

