/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: display.cpp,v 1.8 2003/08/19 17:49:26 rodzilla Exp $

    The CDisplay class handles SDL and OpenGL initialisation as
    well as screenshots.
*/

#include <assert.h>
#include <unistd.h>
#include <SaM3d/config_variable.h>
#include <SaM3d/display.h>
#include <SaM3d/log.h>
#include <SaM3d/opengl.h>
#include <SaM3d/loop.h>
#include "image_transform.h"

#define SCREEN_BPP 24

using namespace std;

namespace SaM3d {

	VideoMode::
	VideoMode(const unsigned int w, const unsigned int h) : m_w(w),
								m_h(h) {
	}

	VideoMode::
	VideoMode(const VideoMode& vm) : m_w(vm.m_w),
					 m_h(vm.m_h) {
	}

	VideoMode&
	VideoMode::
	operator =(const VideoMode& vm) {
		m_w = vm.m_w;
		m_h = vm.m_h;
		return *this;
	}

	bool
	VideoMode::
	operator ==(const VideoMode& vm) const {
		return m_w == vm.m_w && m_h == vm.m_h;
	}

	bool
	VideoMode::
	operator <(const VideoMode& vm) const {
		return m_w < vm.m_w || m_w == vm.m_w && m_h < vm.m_h;
	}





	CDisplay::VideoModeHandler::
	VideoModeHandler(const int max) : m_max(max) { }

	bool
	CDisplay::VideoModeHandler::
	operator ()(float& value) const {
		int tmp = (int)value;

		for(;;) {
			if (tmp > m_max) {
				tmp -= m_max+1;
			} else if (tmp < 0) {
				tmp += m_max+1;
			} else {
				break;
			}
		}
		value = (float)tmp;
		if (CDisplay::m_instance != NULL) {
			CDisplay::m_instance->set_video_mode(tmp);
		}
		return true;
	}





		
	// CDisplay's single instance.
	CDisplay* CDisplay::m_instance = NULL;

	// CDisplay is a singleton. Here's how to obtain the instance.
	CDisplay&
	CDisplay::
	instance() {
		if (m_instance == NULL) {
			m_instance = new CDisplay();
		}
		return *m_instance;
	}

	// The constructor...
	CDisplay::
	CDisplay() : m_video_mode_handler(0),
		     video_mode("video_mode", 0, m_video_mode_handler) {
		// We start with a NULL surface...
		m_surface = NULL;

		// And the default window size is...
		m_window_width = 640;
		m_window_height = 480;

		// Perform video initialization.
		init_display();
	}


	// Video initialization.
	char
	CDisplay::
	init_display() {
		const int buffer_size = 128;
		char buffer[buffer_size+1];

		Log << "Initializing Video\n";

		// Initialize SDL.
		if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0) {
			Log.printf("Video initialization failed: %s\n", SDL_GetError());
			return 1;
		}

		// Whenever the application exits, it'll de-initialize SDL.
		atexit(SDL_Quit);

		Log.printf("Video driver : %s\n", SDL_VideoDriverName(buffer, buffer_size));

		// Create a list of available fullscreen video modes.
		build_video_mode_list();

		set_video_mode(0);

		return 0;
	}

	// Resize the application's rendering area.
	void
	CDisplay::
	resize_window(const unsigned int w, const unsigned int h, const bool windowed) {
		static int video_flags = 0;

		// Calculate the common flags for windowed and fullscreen modes.
		if (video_flags == 0) {
			const SDL_VideoInfo* video_info;

			// Fetch the video info.
			video_info = SDL_GetVideoInfo( );

			if (!video_info) {
				Log.printf("Video query failed: %s\n", SDL_GetError());
				return;
			}

			// The flags to pass to SDL_SetVideoMode.
			video_flags  = SDL_OPENGL;          // Enable OpenGL in SDL.
			video_flags |= SDL_GL_DOUBLEBUFFER; // Enable double buffering.
			video_flags |= SDL_ANYFORMAT;       // Don't emulate requested BPP if not available.
			video_flags |= SDL_HWPALETTE;       // Store the palette in hardware.

			// This checks to see if surfaces can be stored in memory.
			if (video_info->hw_available) {
				Log << "Got Hardware surface.\n";
				video_flags |= SDL_HWSURFACE;
			} else {
				Log << "Got Software surface.\n";
				video_flags |= SDL_SWSURFACE;
			}

			// This checks if hardware blits can be done.
			if (video_info->blit_hw) {
				Log << "Got Hardware blits.\n";
				video_flags |= SDL_HWACCEL;
			}
		}

#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
		OpenGL::clear_context();
#endif

		if (windowed) {
			// If in windowed mode, adjust window size and resize the SDL_Surface.
			m_window_width = w;
			m_window_height = h;

			m_surface = SDL_SetVideoMode(w, h, SCREEN_BPP, video_flags|SDL_RESIZABLE);
		} else {
			// If in fullscreen mode, change the resolution.
			m_surface = SDL_SetVideoMode(w, h, SCREEN_BPP, video_flags|SDL_FULLSCREEN);
		}

		// Adjust viewport position and size.
		m_viewport.move_resize(0, 0, w, h);
		m_viewport.activate();

		if (Loop::M_current_loop != NULL) {
#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
			Loop::M_current_loop->init();
#endif
			Loop::M_current_loop->handle_window_resize(w, h);
		}
	}

	void
	CDisplay::
	set_video_mode(const char _mode) {
		int mode = _mode;

		// If the requested mode is the current mode, return.
		if ((m_instance != NULL) && (mode == video_mode)) return;

		// If mode is invalid, default to 0 (windowed).
		assert ((mode >= 0) && (mode <= m_nb_video_modes));

		// Set the requested mode.
		if (mode == 0) {
			Log.printf("Switching to windowed mode.\n");

			// If windowed mode is requested, use the window mode dimensions.
			resize_window(m_window_width, m_window_height, true);
		} else {
			// If a fullscreen mode is requested, find the resolution in the mode table.
			VideoMode& vid_mode_desc = m_video_modes[mode-1];

			Log.printf("Switching to fullscreen mode #%i (%ix%i).\n",
				   mode, vid_mode_desc.w(), vid_mode_desc.h());
			resize_window(vid_mode_desc.w(), vid_mode_desc.h(), false);
		}

		// Verify there is a surface (i.e. the mode change succeeded).
		if (m_surface == NULL) {
			Log.printf("Video mode set failed: %s\n", SDL_GetError());
			return;
		}

		// If called for the first time (m_instance is NULL) init OpenGL.
		if (m_instance == NULL) {
			OpenGL::init();
		}

		// Set the double buffer OpenGL attribute.
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	}

	const char
	CDisplay::
	get_nb_video_modes() const {
		return m_nb_video_modes;
	}

	const VideoMode&
	CDisplay::
	get_video_mode_info(const char mode) const {
		assert(mode >= 0 && mode <= m_nb_video_modes);
		return m_video_modes[mode];
	}

	void
	CDisplay::
	build_video_mode_list() {
		// Get available fullscreen/hardware modes
		SDL_Rect** sdl_video_modes = SDL_ListModes(NULL, SDL_FULLSCREEN | SDL_HWSURFACE);

		Log << "Building video modes list.\n";

		if (sdl_video_modes == (SDL_Rect**)0) {		// If no mode available exit.
			Log << "No modes available!\n";
			exit(1);
		} else if (sdl_video_modes == (SDL_Rect**)-1) {	// If all mode available, set
			m_nb_video_modes = -1;			// number of modes to -1.
		} else {
			unsigned i;

			for (i = 0; sdl_video_modes[i]; ++i) {
				unsigned j;
				unsigned nb_modes = m_video_modes.size();
				VideoMode video_mode(sdl_video_modes[i]->w, sdl_video_modes[i]->h);
				bool missing = true;

				for (j = 0; j < nb_modes; ++j) {
					if (video_mode == m_video_modes[j]) {
						missing = false;
						break;
					} else if (video_mode < m_video_modes[j]) {
						unsigned k;

						m_video_modes.push_back(m_video_modes[nb_modes-1]);
						for (k = nb_modes-1; k > j; --k) {
							m_video_modes[k] = m_video_modes[k-1];
						}
						m_video_modes[j] = video_mode;
						missing = false;
						break;
					}
				}
				if (missing) {
					m_video_modes.push_back(video_mode);
				}
			}
			m_nb_video_modes = m_video_modes.size();
			m_video_mode_handler.set_max(m_nb_video_modes);
		}
	}

	// Set the window's title.
	void
	CDisplay::
	set_window_title(const char *text) {
		SDL_WM_SetCaption(text, NULL);
	}

	// Take a screenshot of the application.
	// TODO: It's not a good idea to use a temporary buffer.
	//       We'll have to flip SDL surfaces if we wanna use SDL_image to load
	//       texture. There'll be some code to be reused...
	bool
	CDisplay::
	screenshot(const char *filename) const {
		if (!(m_surface->flags & SDL_OPENGL)) {
			SDL_SaveBMP(m_surface, filename);
		} else {
			// The surface that will be used to save the BMP file.
			SDL_Surface *image;

			// Create an SDL surface to hold the image.
			image = SDL_CreateRGBSurface(SDL_SWSURFACE, m_surface->w, m_surface->h, 24,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
						    0x000000FF, 0x0000FF00, 0x00FF0000, 0
#else
						    0x00FF0000, 0x0000FF00, 0x000000FF, 0
#endif
						   );

			// Return if surface creation failed.
			if (image == NULL) return false;

			// Read the image into the buffer.
			glReadPixels(0, 0, m_surface->w, m_surface->h, GL_RGB, GL_UNSIGNED_BYTE, image->pixels);

			ImageTransform::vertical_flip(image);

			// Save the surface into BMP file.
			SDL_SaveBMP(image, filename);

			// Free the surface.
			SDL_FreeSurface(image);
		}

		return true;
	}

}

