/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: file.cpp,v 1.12 2003/08/26 19:54:56 rodzilla Exp $
*/

#include <unistd.h>
#include <dirent.h>
#include <iostream>
#include <list>
#include <string>
#include <zlib.h>
#include <SDL/SDL_rwops.h>
#include <SaM3d/config.h>
#include <SaM3d/file.h>
#include <SaM3d/log.h>

namespace SaM3d {

	using namespace std;

	namespace {
		const unsigned int buffer_size = 1024;
		char buffer[buffer_size];
	}

	class BaseDir {
		unsigned int m_references;

	protected:
		char* m_dir_name;

		BaseDir();

	public:
		virtual ~BaseDir();

		const char* dir_name() const;

		virtual BaseDir* open_dir(const char* path) { return NULL; }

		virtual void get_file_list(set<string>& file_list, const char* dir);

		virtual SDL_RWops* open_file(const char* f, void*& mem);
		SDL_RWops* open_file(const char* f, char** e, void*& mem);

		void add_ref();
		void rem_ref();
	};

	BaseDir::
	BaseDir() {
		m_dir_name = NULL;
	}

	BaseDir::
	~BaseDir() {
		free(m_dir_name);
	}

	const char*
	BaseDir::
	dir_name() const {
		return m_dir_name;
	}

	void
	BaseDir::
	get_file_list(set<string>& file_list, const char* dir){
	}

	SDL_RWops*
	BaseDir::
	open_file(const char* f, void*& mem) {
		mem = NULL;
		return NULL;
	}

	SDL_RWops*
	BaseDir::
	open_file(const char* f, char** e, void*& mem) {
		if (e == NULL) {
			return open_file(f, mem);
		} else {
			char** i;

			for (i = e; i[0] != NULL; ++i) {
				string filename = string(f)+i[0];
				SDL_RWops* rwops;

				rwops = open_file(filename.c_str(), mem);
				if (rwops != NULL) {
					return rwops;
				}
			}
		}

		mem = NULL;
		return NULL;
	}





	class CDirManager {
		list<BaseDir*>m_data[16];

		static unsigned hash(const char* s);

	public:
		CDirManager();
		~CDirManager();

		BaseDir* find(const char* dir_name);
		void add(BaseDir* elem);
		void remove(BaseDir* elem);

		static CDirManager& instance();
	};

#define DirManager (CDirManager::instance())

	CDirManager::
	CDirManager() {
	}

	CDirManager::
	~CDirManager() {
	}

	unsigned
	CDirManager::
	hash(const char* s) {
		unsigned i;
		unsigned rv = 0;

		for (i = 0; s[i] != '\0'; ++i) {
			rv = 31*rv + s[i];
		}

		return rv;
	}

	BaseDir*
	CDirManager::
	find(const char* dir_name) {
		unsigned h = hash(dir_name) % 16;
		list<BaseDir*>::iterator i;

		for (i = m_data[h].begin(); i != m_data[h].end(); ++i) {
			if (!strcmp((*i)->dir_name(), dir_name)) {
				return *i;
			}
		}

		return NULL;
	}

	void
	CDirManager::
	add(BaseDir* elem) {
		if (elem != NULL) {
			unsigned h = hash(elem->dir_name()) % 16;
			{	// Check it's not already registered.
				list<BaseDir*>::iterator i;

				for (i = m_data[h].begin(); i != m_data[h].end(); ++i) {
					if (elem == *i) {
						return;
					}
				}
			}
			m_data[h].push_back(elem);
		}
	}

	void
	CDirManager::
	remove(BaseDir* elem) {
		if (elem != NULL) {
			unsigned h = hash(elem->dir_name()) % 16;
			list<BaseDir*>::iterator i;

			for (i = m_data[h].begin(); i != m_data[h].end();) {
				if (elem == *i) {
					i = m_data[h].erase(i);
				} else {
					++i;
				}
			}
		}
	}

	CDirManager&
	CDirManager::
	instance() {
		static CDirManager i;

		return i;
	}
	








	class Directory : public BaseDir {
		Directory() { }

	public:
		Directory(const char* dir_name);
		virtual ~Directory();

		virtual BaseDir* open_dir(const char* path);

		virtual void get_file_list(set<string>& file_list, const char* dir);

		virtual SDL_RWops* open_file(const char* f, void*& mem);
	};

	class ZipFile : public BaseDir {
	public:
		class Entry {
			string		m_name;
			unsigned int	m_data_position;
			unsigned int	m_size;
			unsigned int	m_compressed_size;
			unsigned short	m_compression_method;
			unsigned int	m_crc;

		public:
			Entry(File& z);

			const string& name() const { return m_name; }
			SDL_RWops* open_file(File& z, void*& mem) const;
		};

	protected:
		File		m_file;
		bool		m_is_correct;

		list<Entry*>	m_entries;
		unsigned short	m_nb_entries;
		unsigned short	m_comment_size;
		unsigned int	m_central_dir_pos;
		unsigned int	m_central_dir_size;
		unsigned int	m_central_dir_offset;

		unsigned int search_central_dir();

		void create_entry_list();

	public:
		ZipFile(const char* file_name);
		~ZipFile();

		virtual void get_file_list(set<string>& file_list, const char* dir);

		virtual SDL_RWops* open_file(const char* f, void*& mem);

	friend class Entry;
	};



	Directory::
	Directory(const char* path) {
		char* current_dir = strdup(getcwd(buffer, buffer_size-1));

		if (chdir(path)) {
			m_dir_name = NULL;
		} else {
			m_dir_name = strdup(getcwd(buffer, buffer_size-1));
			chdir(current_dir);
		}
		free(current_dir);
	}

	Directory::
	~Directory() {
	}

	BaseDir*
	Directory::
	open_dir(const char* path) {
		char* current_dir = strdup(getcwd(buffer, buffer_size-1));
		BaseDir* dir = NULL;

		chdir(m_dir_name);
		{
			Directory* d = new Directory(path);

			if (d->dir_name() != NULL) {
				dir = DirManager.find(d->dir_name());

				if (dir == NULL) {
					DirManager.add(d);
					dir = d;
				} else {
					delete d;
				}
			} else {
				delete d;
			}
		}
		if (dir == NULL) {
			ZipFile* d = new ZipFile(path);

			if (d->dir_name() != NULL) {
				DirManager.add(d);
				dir = d;
			} else {
				delete d;
			}
		}
		chdir(current_dir);
		free(current_dir);

		return dir;
	}

	void
	Directory::
	get_file_list(set<string>& file_list, const char* dir) {
		string path = string(m_dir_name) + "/" + dir;
		DIR *d = opendir(path.c_str());
		struct dirent *entry;
		bool is_dot = !strcmp(dir, ".");

		while ((entry = readdir(d)) != NULL) {
			if (is_dot) {
				file_list.insert(entry->d_name);
			} else {
				file_list.insert(string(dir)+"/"+entry->d_name);
			}
		}
		closedir(d);
	}

	/// FIXME !!!!!
	SDL_RWops*
	Directory::
	open_file(const char* f, void*& mem) {
		if (m_dir_name == NULL) {
			return BaseDir::open_file(f, mem);
		} else {
			char* current_dir = strdup(getcwd(buffer, buffer_size-1));
			SDL_RWops* rwops;

			chdir(m_dir_name);
			rwops = SDL_RWFromFile(f, "rb");
			chdir(current_dir);
			free(current_dir);

			mem = NULL;
			return rwops;
		}
	}


	ZipFile::
	ZipFile(const char* filename) : m_file(filename) {
		{
			char* current_dir = strdup(getcwd(buffer, buffer_size-1));
			char* dir_name = SaM3d::dirname((char*)filename);

			chdir(dir_name);
			free(dir_name);
			getcwd(buffer, buffer_size-1);
			chdir(current_dir);
			free(current_dir);

			strcat(buffer, "/");
			strcat(buffer, basename((char*)filename));
			m_dir_name = strdup(buffer);
		}

		m_is_correct = false;

		unsigned int ul;
		unsigned short number_disk;
		unsigned short number_disk_with_CD;
		unsigned short number_entry_CD;

		m_central_dir_pos = search_central_dir();

		if (m_central_dir_pos == 0) {
			cout << "Zip error 1" << endl;
			return;
		}

		m_file.seek(m_central_dir_pos, SEEK_SET);

		m_file.read_le(ul);
		m_file.read_le(number_disk);
		m_file.read_le(number_disk_with_CD);
		m_file.read_le(m_nb_entries);
		m_file.read_le(number_entry_CD);

		if (   (number_entry_CD != m_nb_entries)
		    || (number_disk_with_CD != 0)
		    || (number_disk != 0)) {
			cout << "Zip error 2" << endl;
			return;
			return;
		}

		m_file.read_le(m_central_dir_size);
		m_file.read_le(m_central_dir_offset);
		m_file.read_le(m_comment_size);

		if (m_central_dir_pos < m_central_dir_offset + m_central_dir_size) {
			cout << "Zip error 3" << endl;
			return;
		}

		m_is_correct = true;

		create_entry_list();
	}

	ZipFile::
	~ZipFile() {
	}

#define COMMENT_BUFFER_SIZE (0x400)

	unsigned int
	ZipFile::
	search_central_dir() {
		char* buffer;
		unsigned long file_size;

		if (m_file.seek(0, SEEK_END) == -1) {
			return 0;
		}
		file_size = m_file.seek(0, SEEK_CUR);

		unsigned long max_global_comment_size = 0xffff;

		if (max_global_comment_size > file_size) {
			max_global_comment_size = file_size;
		}

		buffer = new char[COMMENT_BUFFER_SIZE+4];

		unsigned long back_read;
		unsigned long pos_found = 0;

		for (back_read = 4; back_read < max_global_comment_size; ) {
			if (back_read + COMMENT_BUFFER_SIZE > max_global_comment_size) {
				back_read = max_global_comment_size;
			} else {
				back_read += COMMENT_BUFFER_SIZE;
			}

			unsigned long read_pos = file_size - back_read;
			unsigned long read_size =   (COMMENT_BUFFER_SIZE + 4 < back_read)
						  ? (COMMENT_BUFFER_SIZE + 4)
						  : back_read;

			if (   (m_file.seek(read_pos, SEEK_SET) == -1)
			    || (m_file.read(buffer, read_size, 1) == -1)) {
				break;
			}

			int i;

			for (i = read_size - 3; i >= 0; i--) {
				if (   (buffer[i] == 0x50)
				    && (buffer[i+1] == 0x4b)
				    && (buffer[i+2] == 0x05)
				    && (buffer[i+3] == 0x06)) {
					pos_found = read_pos+i;
					break;
				}
			}
			
			if (pos_found != 0) {
				break;
			}
		}

		delete [] buffer;

		return pos_found;
	}

	void
	ZipFile::
	create_entry_list() {
		unsigned short i;

		m_file.seek(m_central_dir_pos - m_central_dir_size, SEEK_SET);

		for (i = 0; i < m_nb_entries; i++) {
			Entry* e = new Entry(m_file);
			m_entries.push_back(e);
		}
	}

	void
	ZipFile::
	get_file_list(set<string>& file_list, const char* dir) {
		list<Entry*>::iterator i;

		for (i = m_entries.begin(); i != m_entries.end(); ++i) {
			char* dir_name = SaM3d::dirname(((*i)->name().c_str()));

			if (!strcasecmp(dir_name, dir)) {
				file_list.insert(string((*i)->name()));
			}
			free(dir_name);
		}
	}

	SDL_RWops*
	ZipFile::
	open_file(const char* f, void*& mem) {
		list<Entry*>::iterator i;

		for (i = m_entries.begin(); i != m_entries.end(); ++i) {
			if (!strcasecmp((*i)->name().c_str(), f)) {
				return (*i)->open_file(m_file, mem);
			}
		}
		return BaseDir::open_file(f, mem);
	}

	ZipFile::Entry::
	Entry(File& z) {
		unsigned int ul;
		unsigned short us;
		unsigned short size_filename;
		unsigned short size_file_extra;
		unsigned short size_file_comment;
		unsigned int local_file_header_position;
		unsigned int next_central_file_header_position;
		string tmp;

		// Check the central file header signature.
		z.read_le(ul);
		if (ul != 0x02014b50) {
			cout << "Bad magic number" << endl;
		}

		// Get info from the central file header.
		z.read_le(ul);
		z.read_le(m_compression_method);
		z.read_le(us);
		z.read_le(ul);
		z.read_le(m_crc);
		z.read_le(m_compressed_size);
		z.read_le(m_size);
		z.read_le(size_filename);
		z.read_le(size_file_extra);
		z.read_le(size_file_comment);
		z.read_le(ul);
		z.read_le(ul);
		z.read_le(local_file_header_position);
		z.read(m_name, size_filename);
		z.read(tmp, size_file_extra);
		z.read(tmp, size_file_comment);

		// Store the current position (next central file header).
		next_central_file_header_position = z.seek(0, SEEK_CUR);

		// Jump to the local file header.
		z.seek(local_file_header_position, SEEK_SET);

		// Check the local file header signature.
		z.read_le(ul);
		if (ul != 0x04034b50) {
			cout << "Bad magic number" << endl;
		}

		// Skip the local file header.
		// TODO : check against central file header.
		z.read_le(us);
		z.read_le(us);
		z.read_le(us);
		z.read_le(us);
		z.read_le(us);
		z.read_le(ul);
		z.read_le(ul);
		z.read_le(ul);
		z.read_le(us);
		z.read_le(size_file_extra);
		z.read(tmp, size_filename);
		z.read(tmp, size_file_extra);
		// This is where the file data is stored !!!
		m_data_position = z.seek(0, SEEK_CUR);

		// Back to the next central file header position.
		z.seek(next_central_file_header_position, SEEK_SET);
	}

	SDL_RWops*
	ZipFile::Entry::
	open_file(File& z, void*& mem) const {
		int status;

		mem = malloc(m_size);
		z.seek(m_data_position, SEEK_SET);

		if (m_size == m_compressed_size) {
			z.read(mem, m_size, 1);
		} else {
			z_stream_s zs;
			static Bytef* buffer = NULL;
			static unsigned int buffer_size = 0;

			if (buffer_size < m_compressed_size) {
				free(buffer);
				buffer_size = m_compressed_size;
				buffer = (Bytef*)malloc(buffer_size);
			}

			zs.zalloc = (alloc_func)0;
			zs.zfree = (free_func)0;
			zs.opaque = (voidpf)0; 

			if (inflateInit2(&zs, -MAX_WBITS) != Z_OK) {
				return SDL_RWFromMem(NULL, 0);
			}

			zs.next_in = buffer;
			z.read(zs.next_in, m_compressed_size, 1);
			zs.avail_in = buffer_size;
			zs.total_in = 0;
			zs.next_out = (Bytef*)mem;
			zs.avail_out = m_size;
			zs.total_out = 0;

			if ((status = inflate(&zs, Z_FINISH)) != Z_STREAM_END) {
				return SDL_RWFromMem(NULL, 0);
			}

			inflateEnd(&zs);
		}

		return SDL_RWFromMem(mem, m_size);
	}







	BaseDir*
	Dir::
	open_dir(BaseDir* dir, const char* path) {
		if (dir == NULL) {
			{
				Directory* d = new Directory(path);

				if (d->dir_name() != NULL) {
					BaseDir* rv = DirManager.find(d->dir_name());

					if (rv == NULL) {
						DirManager.add(d);

						return d;
					} else {
						delete d;

						return rv;
					}
				} else {
					delete d;
				}
			}

			{
				ZipFile* d = new ZipFile(path);

				if (d->dir_name() != NULL) {
					DirManager.add(d);

					return d;
				} else {
					delete d;
				}
			}

			return NULL;
		} else {
			return dir->open_dir(path);
		}
	}

	Dir::
	Dir(const char* path) {
		char* pathcpy = strdup(path);
                char* tmp;
 
                for (tmp = pathcpy; tmp != NULL;) {
			BaseDir* dir = open_dir(NULL, strsep(&tmp, ";"));

			if (dir) {
				m_dirs.insert(dir);
			}
		}

		free(pathcpy);
	}

	Dir::
	Dir(const Dir& src, const char* path) {
		char* pathcpy = strdup(path);
                char* tmp;
 
                for (tmp = pathcpy; tmp != NULL;) {
			char* dir_name = strsep(&tmp, ";");
			set<BaseDir*>::iterator i;

			for (i = src.m_dirs.begin(); i != src.m_dirs.end(); ++i) {
				BaseDir* dir = open_dir(*i, dir_name);

				if (dir) {
					m_dirs.insert(dir);
				}
			}
		}
	}

	Dir::
	Dir(const Dir& src) {
		m_dirs = src.m_dirs;
	}

	Dir::
	~Dir() {
	}

	Dir&
	Dir::
	operator =(const Dir& src) {
		m_dirs = src.m_dirs;

		return *this;
	}

	bool
	Dir::
	operator ==(const Dir& src) const {
		return m_dirs == src.m_dirs;
	}

	void
	Dir::
	get_file_list(set<string>& file_list, const char* dir) const {
		set<BaseDir*>::iterator i;

		for (i = m_dirs.begin(); i != m_dirs.end(); ++i) {
			if (dir == NULL) {
				(*i)->get_file_list(file_list, ".");
			} else {
				(*i)->get_file_list(file_list, dir);
			}
		}
	}






	File::
	File(const Dir& d, const char* s, char** e) {
		set<BaseDir*>::iterator i;

		for (i = d.m_dirs.begin(); i != d.m_dirs.end(); ++i) {
			if (*i != NULL) {
				m_rwops = (*i)->open_file(s, e, m_mem);
				if (m_rwops != NULL) break;
			}
		}
		m_eof = false;
	}

	File::
	File(const char* s, char** e) {
		Dir d(".");
		set<BaseDir*>::iterator i;

		for (i = d.m_dirs.begin(); i != d.m_dirs.end(); ++i) {
			if (*i != NULL) {
				m_rwops = (*i)->open_file(s, e, m_mem);
				break;
			}
		}
		m_eof = false;
	}

	File::
	~File() {
		free(m_mem);
		if (m_rwops != NULL) {
			SDL_RWclose(m_rwops);
		}
	}

	int
	File::
	seek(int offset, int whence) {
		if (m_rwops != NULL) {
			return SDL_RWseek(m_rwops, offset, whence);
		} else {
			return -1;
		}
	}

	int
	File::
	read(void *ptr, int size, int maxnum) {
		if (m_rwops != NULL) {
			int l = SDL_RWread(m_rwops, ptr, size, maxnum);
			if (l < size*maxnum) {
				m_eof = true;
			}
			return l;
		} else {
			return -1;
		}
	}

	char
	File::
	get() {
		char x;

		read(&x, 1, 1);
		return x;
	}

	bool
	File::
	eof() {
		return m_eof;
	}

	void
	File::
	read(unsigned char& x) {
		read(&x, 1, 1);
	}

	void
	File::
	read_le(unsigned short& x) {
		read(&x, 2, 1);
	}

	void
	File::
	read_le(unsigned int& x) {
		read(&x, 4, 1);
	}

	void
	File::
	read(string& s, unsigned int l) {
		char* buf = new char[l+1];

		read(buf, l, 1);
		buf[l] = '\0';
		s = buf;
		delete[] buf;
	}

}

