/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: file_texture.cpp,v 1.8 2003/08/31 10:04:34 rodzilla Exp $
*/

#include <SaM3d/file_texture.h>
#include <SaM3d/log.h>
#include <SDL/SDL_image.h>
#include "image_transform.h"

//#define SAM3D_DEBUG
#include <SaM3d/debug.h>

namespace SaM3d {

	using namespace std;

	FileTexture::
	FileTexture(const Dir& dir,
		    const char* filename,
		    char** extensions,
		    bool mipmap,
		    GLenum texture_wrap) : m_dir(dir),
					     m_filename(filename),
					     m_extensions(extensions),
					     m_mipmap(mipmap),
					     m_texture_wrap(texture_wrap) {
		DEBUG_FUNCTION("FileTexture::FileTexture(dir,filename,extensions,mipmap,texture_wrap)");
	}

	FileTexture::
	FileTexture(const char* filename,
		    char** extensions,
		    bool mipmap,
		    GLenum texture_wrap) : m_dir("."),
					     m_filename(filename),
					     m_extensions(extensions),
					     m_mipmap(mipmap),
					     m_texture_wrap(texture_wrap) {
		DEBUG_FUNCTION("FileTexture::FileTexture(filename,extensions,mipmap,texture_wrap)");
	}

	FileTexture::
	~FileTexture() {
		DEBUG_FUNCTION("FileTexture::~FileTexture()");
	}

	bool
	FileTexture::
	equal(const SaM3d::BaseTexture* t) const {
		DEBUG_FUNCTION("FileTexture::equal(t)");
		FileTexture* ft;

		return (   (t != NULL)
			&& (t->is_a() == id_file_texture)
			&& (m_dir == (ft = (FileTexture*)t)->m_dir)
			&& (m_extensions == ft->m_extensions)
			&& (m_mipmap == ft->m_mipmap)
			&& (m_texture_wrap == ft->m_texture_wrap)
			&& (m_filename == ft->m_filename));
	}

	unsigned char
	FileTexture::
	hash() const {
		DEBUG_FUNCTION("FileTexture::hash()");
		unsigned char rv = 0;
		unsigned n = m_filename.size();
		unsigned i;

		for (i = 0; i < n; ++i) {
		    rv = 31*rv + m_filename[i];
		}

		return rv;
	}

	void
	FileTexture::
	create() {
		DEBUG_FUNCTION("FileTexture::create()");
		if (m_status == uninitialized) {
			SDL_Surface* image;
			GLenum format;

			{
				File image_file(m_dir, m_filename.c_str(), m_extensions);

				if (image_file.RWops() == NULL) {
					Log << "Cannot open texture file \"" << m_filename << "\".\n";
					m_status = invalid;
					return;
				}
				image = IMG_Load_RW(image_file.RWops(), 0);
				// A quick workaround for the IMG_Load_RW problem with TGA's.
				if (image == NULL) {
					image_file.seek(0, SEEK_SET);
					image = IMG_LoadTyped_RW(image_file.RWops(), 0, "TGA");
				}
			}

			if (image == NULL) {
				Log << "Incorrect texture file \"" << m_filename << "\".\n";
				m_status = invalid;
				return;
			}
			format = ImageTransform::force_supported_format(image);
			ImageTransform::vertical_flip(image);
			ImageTransform::fix_size(image);

			glGenTextures(1, &m_id);
			glBindTexture(GL_TEXTURE_2D, m_id);

			if (m_mipmap) {
				gluBuild2DMipmaps(GL_TEXTURE_2D, image->format->BytesPerPixel,
						  image->w, image->h, format,
						  GL_UNSIGNED_BYTE, image->pixels);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			} else {
				glTexImage2D(GL_TEXTURE_2D, 0, image->format->BytesPerPixel,
					     image->w, image->h, 0, format,
					     GL_UNSIGNED_BYTE, image->pixels);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			}

			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, m_texture_wrap);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, m_texture_wrap);

			SDL_FreeSurface(image);

			m_status = ok;

			CHECK_OPENGL_ERROR;
		}
	}

}

