/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: loop.cpp,v 1.5 2003/08/09 09:35:15 rodzilla Exp $

    The Loop class is a loop skeleton that renders nothing and
    does not react to events.
    Create a subclass of loop to obtain useful loops.
*/

#include <SaM3d/display.h>
#include <SaM3d/log.h>
#include <SaM3d/loop.h>
#include <SaM3d/performance_counter.h>

namespace SaM3d {

	Loop* Loop::M_current_loop = NULL;
	Uint32 Loop::M_current_time = 0;

	// This is the only interesting mehod in this file...
	// It defined the behaviour of the loop.
	// TODO: - recode the "pause" feature.
	//       - support "nested" loops.
	int
	Loop::
	start() {
		int time;			// The time used to calculate frame rendering time.
		bool active;			// True if the loop is active, false if the loop is paused.
		Loop* loop_save;		// The currently running loop.
		static unsigned int center_x, center_y;

		loop_save = M_current_loop;
		M_current_loop = this;
		active = true;
		M_current_time = time = SDL_GetTicks();

		// TODO: Maybe invoking handle_window_resize() here would be a good idea.
		//       Must think about it.
		
		// Store the center coordinates.
		center_x = Display.viewport().w() >> 1;
		center_y = Display.viewport().h() >> 1;
		
		// Invoke the init() method (the one that is supposed to perform OpenGL initialization
		// for this loop).
		init();

		// Hide the mouse cursor and place it somewhere on the window.
		SDL_ShowCursor(SDL_DISABLE);
		SDL_WarpMouse(center_x, center_y);	// TODO: "somewhere" should be at the center of the window.

		// m_done is set to true by the exit() method.
		// set it to false now.
		m_done = false;

		// This is the main loop. It will exit whenever exit() is invoked.
		while (!m_done) {
			SDL_Event event;		// The received events will be stored here.

			bool mouse_motion = false;	// Will be set to true whenever the mouse has to be
							// warped to its initial position.

			bool active_before = active;	// Was the loop active during last frame ?

			// Let's handle all events in the queue !
			while (SDL_PollEvent(&event)) {
				switch (event.type) {
					// On window resize, resize the display (so that the SDL_Surface
					// and the corresponding viewports are updated), and invoke
					// the handle_window_resize method.
					case SDL_VIDEORESIZE:
						Display.resize_window(event.resize.w, event.resize.h, true);
						break;

					// If the application gains focus, activate the loop and hide
					// the mouse cursor, if it looses focus pause the loop and
					// show the cursor.
					case SDL_ACTIVEEVENT:
						if (event.active.state == SDL_APPMOUSEFOCUS) break;
						if (event.active.gain == 0) {
							active = false;
							SDL_ShowCursor(SDL_ENABLE);
						} else {
							active = true;
							SDL_ShowCursor(SDL_DISABLE);
							mouse_motion = true;
						}
						break;			    

					// Handle key presses.
					case SDL_KEYDOWN:
						handle_key_pressed(&event.key.keysym);
						break;

					// All key releases are handled by handle_key_released().
					case SDL_KEYUP:
						handle_key_released(&event.key.keysym);
						break;

					// All mouse button presses are handled by handle_mouse_button_pressed().
					case SDL_MOUSEBUTTONDOWN:
						handle_mouse_button_pressed(event.button.button);
						break;

					// All mouse button releases are handled by handle_mouse_button_pressed().
					case SDL_MOUSEBUTTONUP:
						handle_mouse_button_released(event.button.button);
						break;

					// Mouse motion is handled by handle_mouse_button_pressed().
					case SDL_MOUSEMOTION:
						// If the loop is paused, the event is discarded.
						if (!active) break;

						// If there is an event but the mouse didn't move, do nothing.
						// Not sure this can happen, but I cannot test on all platforms
						// anyway.
						if (event.motion.x == center_x && event.motion.y == center_y) break;

						handle_mouse_motion(event.motion.x-center_x,event.motion.y-center_y);
						mouse_motion = true;
						break;

					// If we receive an SDL_QUIT event (not sure when this can happen), exit.
					case SDL_QUIT:
						exit();
						break;

					default:
						break;
				}
			}

			// Render the scene.
			draw();

			// Update the center coordinates.
			center_x = Display.viewport().w() >> 1;
			center_y = Display.viewport().h() >> 1;

			// Place the mouse cursor somewhere in the window if it moved.
			if (active_before && mouse_motion) {
				SDL_WarpMouse(center_x, center_y);
			}

			// Calculate the time elapsed since the previous frame.
			{
				Uint32 deltaTime;

				M_current_time = SDL_GetTicks();
				deltaTime = M_current_time - time;
				PerformanceCounter.add_frame(time = M_current_time);

				// If the loop is not paused, move the scene.
				if (active) {
					move(deltaTime/1000.0);
				}
			}

			// swap the video buffers.
			SDL_GL_SwapBuffers();
		}

		M_current_loop = loop_save;

		return m_exit_status;
	}

	/* Exit the loop */
	void
	Loop::
	exit(int exit_status) {
		m_exit_status = exit_status;
		m_done = true;
	}

	/* To be redefined in a subclass */
	char
	Loop::
	key_state(const SDLKey k) {
		static Uint8 *keystate = SDL_GetKeyState(NULL);

		return keystate[k] ? 1 : 0;
	}

	char
	Loop::
	mouse_button_state(const Uint8 b) {
		return SDL_GetMouseState(NULL, NULL) & SDL_BUTTON(b);
	}

	/* To be redefined in a subclass */
	void
	Loop::
	init() {
	}

	/* To be redefined in a subclass */
	void
	Loop::
	draw() {
	}

	/* To be redefined in a subclass */
	void
	Loop::
	move(float dt) {
	}

	/* To be redefined in a subclass */
	bool
	Loop::
	handle_key_pressed(const SDL_keysym *keysym) {
		switch (keysym->sym) {
			// Hitting ESCAPE exits the loop.
			case SDLK_ESCAPE:
				exit();
				break;

			// Hitting F10 takes a screenshot.
			case SDLK_F1: {
				Display.video_mode = Display.video_mode-1;
			}	break;

			case SDLK_F2: {
				Display.video_mode = Display.video_mode+1;
			}	break;

			// Hitting F10 takes a screenshot.
			case SDLK_F12:
				Display.screenshot("shot.bmp");
				break;

			default:
				return false;
		}
	
		return true;
	}

	/* To be redefined in a subclass */
	bool
	Loop::
	handle_key_released(const SDL_keysym *keysym) {
		return false;
	}

	/* To be redefined in a subclass */
	bool
	Loop::
	handle_mouse_button_pressed(const Uint8 button) {
		return false;
	}

	/* To be redefined in a subclass */
	bool
	Loop::
	handle_mouse_button_released(const Uint8 button) {
		return false;
	}

	/* To be redefined in a subclass */
	bool
	Loop::
	handle_mouse_motion(const int x, const int y) {
		return false;
	}

	/* To be redefined in a subclass */
	bool
	Loop::
	handle_window_resize(const int w, const int h) {
		return false;
	}

}

