/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: math.cpp,v 1.3 2003/08/07 04:11:03 rodzilla Exp $

    Math-related classes. Vectors, Matrices, Quaternions, Frustum.
*/

#include <SaM3d/math.h>

namespace SaM3d {

	// Call glVertex with the vector's components.
	void
	Vector3::glVertex() const {
		glVertex3f(x, y, z);
	}

	// Call glNormal with the vector's components.
	void
	Vector3::glNormal() const {
		glNormal3f(x, y, z);
	}

	// Call glTranslate with the vector's components.
	void
	Vector3::glTranslate() const {
		glTranslatef(x, y, z);
	}

	// Call glTranslate with the vector's opposite components.
	void
	Vector3::inv_glTranslate() const {
		glTranslatef(-x, -y, -z);
	}

	Plane::
	Plane() : a(0),
		  b(0),
		  c(0),
		  d(0) {
	}

	Plane::
	Plane(const Plane& p) : a(p.a),
				b(p.b),
				c(p.c),
				d(p.d) {
	}

	Plane&
	Plane::
	operator =(const Plane& p) {
		a = p.a;
		b = p.b;
		c = p.c;
		d = p.d;

		return *this;
	}

	bool
	Plane::
	front(const Vector3& v) const {
		return (a*v.x+b*v.y+c*v.z >= d);
	}

	// Copy constructor.
	Matrix3x3::
	Matrix3x3(const Matrix3x3& m) {
		m_data[0]  = m.m_data[0];
		m_data[1]  = m.m_data[1];
		m_data[2]  = m.m_data[2];
		m_data[3]  = m.m_data[3];
		m_data[4]  = m.m_data[4];
		m_data[5]  = m.m_data[5];
		m_data[6]  = m.m_data[6];
		m_data[7]  = m.m_data[7];
		m_data[8]  = m.m_data[8];
	}

	// Create a Matrix3x3 from a quaternion.
	Matrix3x3::
	Matrix3x3(const Quaternion& q) {
		m_data[0] = 1.0f - 2.0f * ( q.m_y * q.m_y + q.m_z * q.m_z );  
		m_data[1] = 2.0f * ( q.m_x * q.m_y - q.m_w * q.m_z );  
		m_data[2] = 2.0f * ( q.m_x * q.m_z + q.m_w * q.m_y );  
		m_data[3] = 2.0f * ( q.m_x * q.m_y + q.m_w * q.m_z );  
		m_data[4] = 1.0f - 2.0f * ( q.m_x * q.m_x + q.m_z * q.m_z );  
		m_data[5] = 2.0f * ( q.m_y * q.m_z - q.m_w * q.m_x );  
		m_data[6] = 2.0f * ( q.m_x * q.m_z - q.m_w * q.m_y );  
		m_data[7] = 2.0f * ( q.m_y * q.m_z + q.m_w * q.m_x );  
		m_data[8] = 1.0f - 2.0f * ( q.m_x * q.m_x + q.m_y * q.m_y );  
	}

	Vector3
	Matrix3x3::
	operator *(const Vector3& op) const {
		Vector3 result;

		result.x = op.x*m_data[0]+op.y*m_data[3]+op.z*m_data[6];
		result.y = op.x*m_data[1]+op.y*m_data[4]+op.z*m_data[7];
		result.z = op.x*m_data[2]+op.y*m_data[5]+op.z*m_data[8];

		return result;
	}

	Quaternion::
	Quaternion() {
		m_w = 1; m_x = m_y = m_z = 0;
	}
 
	Quaternion::
	Quaternion(const Vector3& v, const float a) {
		float sha = sin(0.5*a)/v.length();

		m_w = cos(0.5*a);
		m_x = v.x*sha;
		m_y = v.y*sha;
		m_z = v.z*sha;
	}

	Quaternion::
	Quaternion::
	Quaternion(const Quaternion& s) {
		m_w = s.m_w;
		m_x = s.m_x;
		m_y = s.m_y;
		m_z = s.m_z;
	}

	Quaternion&
	Quaternion::
	operator =(const Quaternion& q) {
		m_w = q.m_w;
		m_x = q.m_x;
		m_y = q.m_y;
		m_z = q.m_z;

		return *this;
	}

	Quaternion&
	Quaternion::
	operator *=(const Quaternion& op) {
		float w = m_w;
		float x = m_x;
		float y = m_y;
		float z = m_z;

		m_w = w*op.m_w - x*op.m_x - y*op.m_y - z*op.m_z;
		m_x = w*op.m_x + x*op.m_w + y*op.m_z - z*op.m_y;
		m_y = w*op.m_y + y*op.m_w + z*op.m_x - x*op.m_z;
		m_z = w*op.m_z + z*op.m_w + x*op.m_y - y*op.m_x;

		return *this;
	}

	Quaternion
	Quaternion::
	operator -() const {
		Quaternion r;

		r.m_w = m_w;
		r.m_x = -m_x;
		r.m_y = -m_y;
		r.m_z = -m_z;

		return r;
	}

	Quaternion
	Quaternion::
	operator *(const Quaternion& op) const {
		Quaternion r;

		r.m_w = m_w*op.m_w - m_x*op.m_x - m_y*op.m_y - m_z*op.m_z;
		r.m_x = m_w*op.m_x + m_x*op.m_w + m_y*op.m_z - m_z*op.m_y;
		r.m_y = m_w*op.m_y + m_y*op.m_w + m_z*op.m_x - m_x*op.m_z;
		r.m_z = m_w*op.m_z + m_z*op.m_w + m_x*op.m_y - m_y*op.m_x;

		return r;
	}

	void
	Quaternion::
	glRotate() const {
		float	angle;		// temp angle
		float	scale;		// temp vars

		angle = acos(m_w);

		scale = (float)sqrt(m_x*m_x+m_y*m_y+m_z*m_z);

		if (scale != 0.0) {
			glRotatef(114.59155902616*angle, m_x/scale, m_y/scale, m_z/scale);
		}
	}

	void
	Quaternion::
	inv_glRotate() const {
		float	angle;		// temp angle
		float	scale;		// temp vars

		angle = acos(m_w);

		scale = (float)sqrt(m_x*m_x+m_y*m_y+m_z*m_z);

		if (scale != 0.0) {
			glRotatef(-114.59155902616*angle, m_x/scale, m_y/scale, m_z/scale);
		}
	}



	void
	Frustum::
	update() {
	   float   proj[16];
	   float   modl[16];
	   float   clip[16];
	   float   t;

	   /* Get the current PROJECTION matrix from OpenGL */
	   glGetFloatv( GL_PROJECTION_MATRIX, proj );

	   /* Get the current MODELVIEW matrix from OpenGL */
	   glGetFloatv( GL_MODELVIEW_MATRIX, modl );

	   /* Combine the two matrices (multiply projection by modelview) */
	   clip[0] = modl[0] * proj[0] + modl[1] * proj[4] + modl[2] * proj[8] + modl[3] * proj[12];
	   clip[1] = modl[0] * proj[1] + modl[1] * proj[5] + modl[2] * proj[9] + modl[3] * proj[13];
	   clip[2] = modl[0] * proj[2] + modl[1] * proj[6] + modl[2] * proj[10] + modl[3] * proj[14];
	   clip[3] = modl[0] * proj[3] + modl[1] * proj[7] + modl[2] * proj[11] + modl[3] * proj[15];

	   clip[4] = modl[4] * proj[0] + modl[5] * proj[4] + modl[6] * proj[8] + modl[7] * proj[12];
	   clip[5] = modl[4] * proj[1] + modl[5] * proj[5] + modl[6] * proj[9] + modl[7] * proj[13];
	   clip[6] = modl[4] * proj[2] + modl[5] * proj[6] + modl[6] * proj[10] + modl[7] * proj[14];
	   clip[7] = modl[4] * proj[3] + modl[5] * proj[7] + modl[6] * proj[11] + modl[7] * proj[15];

	   clip[8] = modl[8] * proj[0] + modl[9] * proj[4] + modl[10] * proj[8] + modl[11] * proj[12];
	   clip[9] = modl[8] * proj[1] + modl[9] * proj[5] + modl[10] * proj[9] + modl[11] * proj[13];
	   clip[10] = modl[8] * proj[2] + modl[9] * proj[6] + modl[10] * proj[10] + modl[11] * proj[14];
	   clip[11] = modl[8] * proj[3] + modl[9] * proj[7] + modl[10] * proj[11] + modl[11] * proj[15];

	   clip[12] = modl[12] * proj[0] + modl[13] * proj[4] + modl[14] * proj[8] + modl[15] * proj[12];
	   clip[13] = modl[12] * proj[1] + modl[13] * proj[5] + modl[14] * proj[9] + modl[15] * proj[13];
	   clip[14] = modl[12] * proj[2] + modl[13] * proj[6] + modl[14] * proj[10] + modl[15] * proj[14];
	   clip[15] = modl[12] * proj[3] + modl[13] * proj[7] + modl[14] * proj[11] + modl[15] * proj[15];

	   /* Extract the numbers for the RIGHT plane */
	   m_data[0][0] = clip[3] - clip[0];
	   m_data[0][1] = clip[7] - clip[4];
	   m_data[0][2] = clip[11] - clip[8];
	   m_data[0][3] = clip[15] - clip[12];

	   /* Normalize the result */
	   t = sqrt( m_data[0][0] * m_data[0][0] + m_data[0][1] * m_data[0][1] + m_data[0][2] * m_data[0][2] );
	   m_data[0][0] /= t;
	   m_data[0][1] /= t;
	   m_data[0][2] /= t;
	   m_data[0][3] /= t;

	   /* Extract the numbers for the LEFT plane */
	   m_data[1][0] = clip[3] + clip[0];
	   m_data[1][1] = clip[7] + clip[4];
	   m_data[1][2] = clip[11] + clip[8];
	   m_data[1][3] = clip[15] + clip[12];

	   /* Normalize the result */
	   t = sqrt( m_data[1][0] * m_data[1][0] + m_data[1][1] * m_data[1][1] + m_data[1][2] * m_data[1][2] );
	   m_data[1][0] /= t;
	   m_data[1][1] /= t;
	   m_data[1][2] /= t;
	   m_data[1][3] /= t;

	   /* Extract the BOTTOM plane */
	   m_data[2][0] = clip[3] + clip[1];
	   m_data[2][1] = clip[7] + clip[5];
	   m_data[2][2] = clip[11] + clip[9];
	   m_data[2][3] = clip[15] + clip[13];

	   /* Normalize the result */
	   t = sqrt( m_data[2][0] * m_data[2][0] + m_data[2][1] * m_data[2][1] + m_data[2][2] * m_data[2][2] );
	   m_data[2][0] /= t;
	   m_data[2][1] /= t;
	   m_data[2][2] /= t;
	   m_data[2][3] /= t;

	   /* Extract the TOP plane */
	   m_data[3][0] = clip[3] - clip[1];
	   m_data[3][1] = clip[7] - clip[5];
	   m_data[3][2] = clip[11] - clip[9];
	   m_data[3][3] = clip[15] - clip[13];

	   /* Normalize the result */
	   t = sqrt( m_data[3][0] * m_data[3][0] + m_data[3][1] * m_data[3][1] + m_data[3][2] * m_data[3][2] );
	   m_data[3][0] /= t;
	   m_data[3][1] /= t;
	   m_data[3][2] /= t;
	   m_data[3][3] /= t;

	   /* Extract the FAR plane */
	   m_data[4][0] = clip[3] - clip[2];
	   m_data[4][1] = clip[7] - clip[6];
	   m_data[4][2] = clip[11] - clip[10];
	   m_data[4][3] = clip[15] - clip[14];

	   /* Normalize the result */
	   t = sqrt( m_data[4][0] * m_data[4][0] + m_data[4][1] * m_data[4][1] + m_data[4][2] * m_data[4][2] );
	   m_data[4][0] /= t;
	   m_data[4][1] /= t;
	   m_data[4][2] /= t;
	   m_data[4][3] /= t;

	   /* Extract the NEAR plane */
	   m_data[5][0] = clip[3] + clip[2];
	   m_data[5][1] = clip[7] + clip[6];
	   m_data[5][2] = clip[11] + clip[10];
	   m_data[5][3] = clip[15] + clip[14];

	   /* Normalize the result */
	   t = sqrt( m_data[5][0] * m_data[5][0] + m_data[5][1] * m_data[5][1] + m_data[5][2] * m_data[5][2] );
	   m_data[5][0] /= t;
	   m_data[5][1] /= t;
	   m_data[5][2] /= t;
	   m_data[5][3] /= t;
	}

	bool
	Frustum::
	test_sphere(const Vector3& center, const float radius) const {
		int p;
		float d;

		for(p = 0; p < 6; p++) {
			d = m_data[p][0]*center.x+m_data[p][1]*center.y+
			    m_data[p][2]*center.z+m_data[p][3];
			if (d <= -radius) {
				return true;
			}
		}
		return false;
	}

	bool
	Frustum::
	test_box(const Vector3& mins, const Vector3& maxs) const {
		unsigned char i;

		for (i = 0; i < 6; i++ ) {
			if(m_data[i][0] * mins.x + m_data[i][1] * mins.y + m_data[i][2] * mins.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * maxs.x + m_data[i][1] * mins.y + m_data[i][2] * mins.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * mins.x + m_data[i][1] * maxs.y + m_data[i][2] * mins.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * maxs.x + m_data[i][1] * maxs.y + m_data[i][2] * mins.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * mins.x + m_data[i][1] * mins.y + m_data[i][2] * maxs.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * maxs.x + m_data[i][1] * mins.y + m_data[i][2] * maxs.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * mins.x + m_data[i][1] * maxs.y + m_data[i][2] * maxs.z + m_data[i][3] > 0)  continue;
			if(m_data[i][0] * maxs.x + m_data[i][1] * maxs.y + m_data[i][2] * maxs.z + m_data[i][3] > 0)  continue;

			// If we get here, it isn't in the frustum
			return false;
		}

		// Return a true for the box being inside of the frustum
		return true;
	}

}

