/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: opengl.cpp,v 1.4 2003/08/31 10:04:34 rodzilla Exp $
*/

#include <SaM3d/config.h>
#include <stdio.h>
#include <string.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <SDL/SDL.h>
#include <SaM3d/log.h>
#include <SaM3d/texture.h>

//#define SAM3D_DEBUG
#include <SaM3d/debug.h>

namespace SaM3d {

#ifdef __WINDOWS__
	PFNGLFOGCOORDFEXTPROC		glFogCoordfEXT;
	PFNGLACTIVETEXTUREPROC		glActiveTexture;
	PFNGLCLIENTACTIVETEXTUREPROC	glClientActiveTexture;
#endif

	namespace OpenGL {

		bool check_extension(const char* extension_name) {
			DEBUG_FUNCTION("OpenGL::check_extension(extension_name)");
			static char* extension_list = (char*)glGetString(GL_EXTENSIONS);
			unsigned int extension_name_length = strlen(extension_name);
			char* tmp = extension_list;

			if (!extension_name || !extension_list) return false;

			while (tmp[0]) {
				unsigned int first_extension_length = strcspn(tmp, " ");

				if (   extension_name_length == first_extension_length
				    && strncmp(extension_name, tmp, first_extension_length) == 0) {
					return true;
				}
				tmp += first_extension_length + 1;
			}

			return false;
		}

		void init () {
			DEBUG_FUNCTION("OpenGL::init()");
			Log.printf("Initializing OpenGL\n");
			Log.printf("Renderer : %s\n", (char *)glGetString(GL_RENDERER));
			Log.printf("Vendor : %s\n", (char *)glGetString(GL_VENDOR));
			Log.printf("Version : %s\n", (char *)glGetString(GL_VERSION));


#ifdef __WINDOWS__
			/*
			glFogCoordfEXT = (PFNGLFOGCOORDFEXTPROC)SDL_GL_GetProcAddress("glFogCoordfEXT");
			if (glFogCoordfEXT == NULL) {
				Log.printf("Cannot find procedure address (glFogCoordfEXT)\n");
			}
			*/
			glActiveTexture = (PFNGLACTIVETEXTUREPROC)SDL_GL_GetProcAddress("glActiveTexture");
			if (glActiveTexture == NULL) {
				Log.printf("Cannot find procedure address (glActiveTexture)\n");
			}

			glClientActiveTexture = (PFNGLCLIENTACTIVETEXTUREPROC)SDL_GL_GetProcAddress("glClientActiveTexture");
			if (glActiveTexture == NULL) {
				Log.printf("Cannot find procedure address (glClientActiveTexture)\n");
			}
#endif

			if (Texture::M_has_multitexture = check_extension("GL_ARB_multitexture")) {
				glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &Texture::M_nb_tex_units);
				Log.printf("Max texture units : %i.\n", Texture::M_nb_tex_units);
			}

			if (Texture::M_bgra_support = check_extension("GL_EXT_bgra")) {
				Log.printf("BGRA format is supported.\n");
			}

		}

#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
	
		void clear_context() {
			DEBUG_FUNCTION("OpenGL::clear_context()");
			Texture::release_all();
		}

#endif

		void alpha_func(GLenum func, float ref) {
			DEBUG_FUNCTION("OpenGL::alpha_func(func,ref)");
			static GLenum last_func = GL_ALWAYS;
			static float last_ref = 0.0;

			if (func == GL_ALWAYS) {
				if (last_func != GL_ALWAYS) {
					glDisable(GL_ALPHA_TEST);
					last_func = GL_ALWAYS;
				}
			} else if (   (func != last_func)
				   || (ref != last_ref)) {
				glEnable(GL_ALPHA_TEST);
				glAlphaFunc(func, ref);
				last_func = func;
				last_ref = ref;
			}
			CHECK_OPENGL_ERROR;
		}

		void cull_face(GLenum face) {
			DEBUG_FUNCTION("OpenGL::cull_face(face)");
			static GLenum last_face = GL_BACK;

			if (face != last_face) {
				if (face == GL_NONE) {
					glDisable(GL_CULL_FACE);
					CHECK_OPENGL_ERROR;
				} else {
					glEnable(GL_CULL_FACE);
					if (face == GL_FRONT) {
						glCullFace(GL_FRONT);
						CHECK_OPENGL_ERROR;
					} else if (face == GL_BACK) {
						glCullFace(GL_BACK);
						CHECK_OPENGL_ERROR;
					} else {
						assert(false && "Error: bad parameter for CULL_FACE.\n");
					}
				}
				last_face = face;
			}
		}

	}

}

