/**************************************************************************

    This file is part of SaM3d!, yet another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: texture.cpp,v 1.8 2003/08/24 15:45:00 rodzilla Exp $
*/

#include <assert.h>
#include <list>
#include <iostream>
#include <SaM3d/log.h>
#include <SaM3d/texture.h>

//#define SAM3D_DEBUG
#include <SaM3d/debug.h>

namespace SaM3d {

	using namespace std;

	namespace {

		template<class T, unsigned char table_size>
		class GenericManager {
		protected:
			list<T*>	m_data[table_size];

		public:
			GenericManager();
			~GenericManager();

			T* find(T* elem);
			void unregister(T* elem);
		};

		template<class T, unsigned char table_size>
		GenericManager<T, table_size>::
		GenericManager() {
			DEBUG_FUNCTION("GenericManager<T, table_size>::GenericManager()");
		}

		template<class T, unsigned char table_size>
		GenericManager<T, table_size>::
		~GenericManager() {
			DEBUG_FUNCTION("~GenericManager<T, table_size>::GenericManager()");
		}

		template<class T, unsigned char table_size>
		T*
		GenericManager<T, table_size>::
		find(T* elem) {
			DEBUG_FUNCTION("~GenericManager<T, table_size>::find(elem)");
			unsigned hash = elem->hash() % table_size;
			typename list<T*>::iterator i;

			for (i = m_data[hash].begin(); i != m_data[hash].end(); ++i) {
				if ((*i)->equal(elem)) {
					if (elem != *i) {
						delete elem;
					}
					return *i;
				}
			}

			m_data[hash].push_back(elem);
			return elem;
		}

		template<class T, unsigned char table_size>
		void
		GenericManager<T, table_size>::
		unregister(T* elem) {
			DEBUG_FUNCTION("~GenericManager<T, table_size>::unregister(elem)");
			if (elem != NULL) {
				unsigned hash = elem->hash() % table_size;
				typename list<T*>::iterator i;

				for (i = m_data[hash].begin(); i != m_data[hash].end();) {
					if (elem == *i) {
						i = m_data[hash].erase(i);
					} else {
						++i;
					}
				}
			}
		}

	}

#define TEX_MGR_HASH_SIZE 32

	class Texture::Manager : public GenericManager<BaseTexture, TEX_MGR_HASH_SIZE> {
	public:
		Manager();

		static Manager& instance();

		BaseTexture* find(BaseTexture* elem);
		void release_all();
	};

	Texture::Manager::
	Manager() : GenericManager<BaseTexture, TEX_MGR_HASH_SIZE>() {
		DEBUG_FUNCTION("Texture::Manager::Manager()");
		find(new BaseTexture());
	}

	Texture::Manager&
	Texture::Manager::
	instance() {
		DEBUG_FUNCTION("Texture::Manager::instance()");
		static Texture::Manager texture_manager;

		return texture_manager;
	}

	BaseTexture*
	Texture::Manager::
	find(BaseTexture* elem) {
		DEBUG_FUNCTION("Texture::Manager::find(elem)");
		if (elem == NULL) {
			return Texture::none();
		} else {
			return GenericManager<BaseTexture, TEX_MGR_HASH_SIZE>::find(elem);
		}
	}

#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
	void
	Texture::Manager::
	release_all() {
		DEBUG_FUNCTION("Texture::Manager::release_all()");
		int i;

		for (i = 0; i < TEX_MGR_HASH_SIZE; ++i) {
			list<BaseTexture*>::iterator j;

			for (j = m_data[i].begin(); j != m_data[i].end(); ++j) {
				(*j)->release();
			}
		}
	}
#endif

#define texture_manager Texture::Manager::instance()



	bool Texture::M_has_multitexture = false;
	GLint Texture::M_nb_tex_units = 1;
	unsigned char Texture::M_texture_unit = 0;
	BaseTexture* Texture::M_current_textures[32] = {
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
	};

	bool Texture::M_bgra_support = false;

	namespace {
		GLint texture_units[] = {
			GL_TEXTURE0, GL_TEXTURE1, GL_TEXTURE2, GL_TEXTURE3,
			GL_TEXTURE4, GL_TEXTURE5, GL_TEXTURE6, GL_TEXTURE7,
			GL_TEXTURE8, GL_TEXTURE9, GL_TEXTURE10, GL_TEXTURE11,
			GL_TEXTURE12, GL_TEXTURE13, GL_TEXTURE14, GL_TEXTURE15,
			GL_TEXTURE16, GL_TEXTURE17, GL_TEXTURE18, GL_TEXTURE19,
			GL_TEXTURE20, GL_TEXTURE21, GL_TEXTURE22, GL_TEXTURE23,
			GL_TEXTURE24, GL_TEXTURE25, GL_TEXTURE26, GL_TEXTURE27,
			GL_TEXTURE28, GL_TEXTURE29, GL_TEXTURE30, GL_TEXTURE31
		};
	}

	void
	Texture::
	set_texture_unit(const unsigned char u) {
		DEBUG_FUNCTION("Texture::set_texture_unit(u)");
		if (u != M_texture_unit) {
			assert(M_has_multitexture);
			assert(u < M_nb_tex_units);
			glActiveTexture(texture_units[u]);
			glClientActiveTexture(texture_units[u]);
			M_texture_unit = u;
		}
	}

	/* Texture class - Sphere mapping */

	void
	Texture::
	tc_gen(const GLenum mode) {
		DEBUG_FUNCTION("Texture::tc_gen(mode)");
		static bool last_mode = GL_NONE;

		if (mode == last_mode) return;

		switch (mode) {
			case GL_OBJECT_LINEAR:
			case GL_EYE_LINEAR:
			case GL_SPHERE_MAP:
				glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, mode);
				glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, mode);	
				glEnable(GL_TEXTURE_GEN_S);
				glEnable(GL_TEXTURE_GEN_T);
				last_mode = mode;
				break;
			default:
				glDisable(GL_TEXTURE_GEN_S);
				glDisable(GL_TEXTURE_GEN_T);
				last_mode = GL_NONE;
				break;
		}
	}

#ifdef RESTORE_OPENGL_CONTEXT_ON_RESOLUTION_CHANGE
	void
	Texture::
	release_all() {
		DEBUG_FUNCTION("Texture::release_all()");
		texture_manager.release_all();
	}
#endif

	const Texture&
	Texture::
	none() {
		DEBUG_FUNCTION("Texture::none()");
		static Texture no_texture(new BaseTexture());
		return no_texture;
	}

	Texture::
	Texture() : SmartPtr<BaseTexture>(texture_manager.find(NULL)) {
		DEBUG_FUNCTION("Texture::Texture(BaseTexture* t)");
	}

	Texture::
	Texture(BaseTexture* t) : SmartPtr<BaseTexture>(texture_manager.find(t)) {
		DEBUG_FUNCTION("Texture::Texture(BaseTexture* t)");
	}

	Texture::
	Texture(const Texture& t) : SmartPtr<BaseTexture>(t) {
		DEBUG_FUNCTION("Texture::Texture(const Texture& t)");
	}

	Texture::
	~Texture() {
		DEBUG_FUNCTION("Texture::&Texture()");
	}

	Texture&
	Texture::
	operator=(BaseTexture* t) {
		DEBUG_FUNCTION("Texture::operator =(t)");
		BaseTexture* tex;
		if (t == NULL) {
			tex = none().m_ptr;
		} else {
			tex = texture_manager.find(t);
		}
		assert(tex != NULL);
		if (tex != m_ptr) {
			m_ptr->rem_ref();
			m_ptr = tex;
			m_ptr->add_ref();
		}
		return *this;
	}




	BaseTexture::
	BaseTexture() : RefCounted(),
			m_status(uninitialized) {
		DEBUG_FUNCTION("BaseTexture::BaseTexture()");
	}

	BaseTexture::
	~BaseTexture() {
		DEBUG_FUNCTION("BaseTexture::~BaseTexture()");
		release();
	}

	void
	BaseTexture::
	free() {
		DEBUG_FUNCTION("BaseTexture::free()");
		texture_manager.unregister(this);
		delete this;
	}

	bool
	BaseTexture::
	equal(const BaseTexture* t) const {
		DEBUG_FUNCTION("BaseTexture::equal(t)");
		return (t == NULL || t->is_a() == id_base_texture);
	}

	unsigned char
	BaseTexture::
	hash() const {
		DEBUG_FUNCTION("BaseTexture::hash()");
		return 0;
	}

	void
	BaseTexture::
	create() {
		DEBUG_FUNCTION("BaseTexture::create()");
		m_status = no_texture;
	}

	void
	BaseTexture::
	release() {
		DEBUG_FUNCTION("BaseTexture::release()");
		if (m_status == ok) {
			glDeleteTextures(1, &m_id);
		}
		m_status = uninitialized;
	}

	void
	BaseTexture::
	bind() {
		DEBUG_FUNCTION("BaseTexture::bind()");
		if (Texture::M_current_textures[Texture::M_texture_unit] != this) {
			create();
			assert(m_status != uninitialized);

			if (m_status == ok) {
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, m_id);
			} else {
				glDisable(GL_TEXTURE_2D);
			}
			Texture::M_current_textures[Texture::M_texture_unit] = this;
		}
	}

}
