/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: bsp.cpp,v 1.5 2003/08/31 10:04:34 rodzilla Exp $
*/

#include <iostream>
#include "bsp.h"

namespace SaM3d {

	using namespace std;

	BSP::Node::node_info_t::
	node_info_t() : m_plane(NULL),
			m_front_node(NULL),
			m_back_node(NULL) {
	}


	BSP::Node::
	Node() : m_node_info(NULL),
		 m_leaf_info(NULL),
		 m_bounding_volume(NULL) {
	}

	BSP::Node::node_info_t&
	BSP::Node::
	get_node_info() {
		assert(m_leaf_info == NULL);

		if (m_node_info == NULL) {
			m_node_info = new node_info_t;
			m_node_info->m_plane = NULL;
			m_node_info->m_front_node = NULL;
			m_node_info->m_back_node = NULL;
		}

		return *m_node_info;
	}

	BSP::Node::leaf_info_t&
	BSP::Node::
	get_leaf_info() {
		assert(m_node_info == NULL);

		if (m_leaf_info == NULL) {
			m_leaf_info = new leaf_info_t;
		}

		return *m_leaf_info;
	}

	void
	BSP::Node::
	set_plane(Plane* p) {
		node_info_t& node_info = get_node_info();

		assert(node_info.m_plane == NULL);
		node_info.m_plane = p;
	}

	void
	BSP::Node::
	set_front_node(Node* n) {
		node_info_t& node_info = get_node_info();

		assert(node_info.m_front_node == NULL);
		node_info.m_front_node = n;
	}

	void
	BSP::Node::
	set_back_node(Node* n) {
		node_info_t& node_info = get_node_info();

		assert(node_info.m_back_node == NULL);
		node_info.m_back_node = n;
	}

	void
	BSP::Node::
	add_visible_leaf(Node* n) {
		leaf_info_t& leaf_info = get_leaf_info();

		leaf_info.m_pvs.push_back(n);

		if (m_bounding_volume && n->m_bounding_volume) {
			Vector3 center = m_bounding_volume->center();
			float l1 = (n->m_bounding_volume->center() - center).length2();
			int i;

			for (i = leaf_info.m_pvs.size()-2; i >= 0; --i) {
				if (leaf_info.m_pvs[i]->m_bounding_volume != NULL) {
					float l2 = (leaf_info.m_pvs[i]->m_bounding_volume->center() - center).length2();
					Node* tmp;

					if (l2 < l1) {
						break;
					}
					tmp = leaf_info.m_pvs[i];
					leaf_info.m_pvs[i] = leaf_info.m_pvs[i+1];
					leaf_info.m_pvs[i+1] = tmp;
				} else {
					break;
				}
			}
		}
	}

	void
	BSP::Node::
	add_mesh(O3d* m) {
			leaf_info_t& leaf_info = get_leaf_info();

			leaf_info.m_meshes.push_back(m);
	}

	void
	BSP::Node::
	set_bounding_volume(BoundingVolume* v) {
		assert(m_bounding_volume == NULL);
		m_bounding_volume = v;
	}

	BSP::Node*
	BSP::Node::
	find_leaf_node(const Vector3& p) {
		Node* n = this;

		do {
			if (n->m_node_info == NULL) {
				return n;
			} else {
				if (n->m_node_info->m_plane == NULL) {
					n = NULL;
				} else if (n->m_node_info->m_plane->front(p)) {
					n = n->m_node_info->m_front_node;
				} else {
					n = n->m_node_info->m_back_node;
				}
			}
		} while (n != NULL);

		return NULL;
	}

	bool
	BSP::Node::
	is_visible(const Frustum& f) const {
		return true;
	}

	vector<BSP::Node*>&
	BSP::Node::
	get_visible_leaves() {
		if (m_leaf_info != NULL) {
			return m_leaf_info->m_pvs;
		} else {
			static std::vector<Node*> dummy;

			return dummy;
		}
	}

	void
	BSP::Node::
	draw(vector<O3d*>& transparent_meshes) {
		if (m_leaf_info != NULL) {
			vector<O3d*>& meshes = m_leaf_info->m_meshes;
			vector<O3d*>::iterator i;

			for (i = meshes.begin(); i != meshes.end(); ++i) {
				if (!(*i)->is_up_to_date()) {
					if ((*i)->is_transparent()) {
						transparent_meshes.push_back(*i);
					} else {
						(*i)->draw();
					}
					(*i)->update();
				}
			}
		}
	}

	void
	BSP::Node::
	draw(const Frustum& f, vector<O3d*>& transparent_meshes) {
		if (m_leaf_info != NULL) {
			vector<O3d*>& meshes = m_leaf_info->m_meshes;
			vector<O3d*>::iterator i;

			if (   (m_bounding_volume != NULL)
			    && (!m_bounding_volume->is_visible(f))) {
				return;
			}

			for (i = meshes.begin(); i != meshes.end(); ++i) {
				if (!(*i)->is_up_to_date()) {
					if ((*i)->is_transparent()) {
						transparent_meshes.push_back(*i);
					} else {
						(*i)->draw(f);
					}
					(*i)->update();
				}
			}
		}
	}

}

