/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: loader_bsp.cpp,v 1.16 2003/08/31 10:04:34 rodzilla Exp $
*/

#include <unistd.h>
#include <iostream>
#include <list>
#include <SaM3d/file_texture.h>
#include <SaM3d/font.h>
#include <SaM3d/lightmap.h>
#include <SaM3d/log.h>
#include <SaM3d/viewport.h>
#include "bsp.h"
#include "loader_bsp.h"
#include "q3shaders.h"

namespace SaM3d {
	
	using namespace std;

	struct bsp_header_t {
	    char str_id[4];
	    int  version;
	};

	struct bsp_lump_t {
		int offset;
		int length;
	};

	struct bsp_lumps_t {
		bsp_lump_t entities;
		bsp_lump_t textures;
		bsp_lump_t planes;
		bsp_lump_t nodes;
		bsp_lump_t leafs;
		bsp_lump_t leaf_faces;
		bsp_lump_t leaf_brushes;
		bsp_lump_t models;
		bsp_lump_t brushes;
		bsp_lump_t brush_sides;
		bsp_lump_t vertices;
		bsp_lump_t mesh_vertices;
		bsp_lump_t effects;
		bsp_lump_t faces;
		bsp_lump_t lightmaps;
		bsp_lump_t light_volumes;
		bsp_lump_t visibility_data;
	};

	struct bsp_texture_t {
		char name[64];
		int flags;
		int contents;
	};

	struct bsp_node_t {
		int plane;
		int children[2];
		int mins[3];
		int maxs[3];
	};

	struct bsp_leaf_t {
		int cluster;
		int area;
		int mins[3];
		int maxs[3];
		int leafface;
		int n_leaffaces;
		int leafbrush;
		int n_leafbrushes;
	};

	struct bsp_face_t {
		int texture;
		int effect;
		int type;
		int first_vertex;
		int nb_vertices;
		int first_mesh_vertex;
		int nb_mesh_vertices;
		int lightmap;	
		int lMapCorner[2];
		int lMapSize[2];
		Vector3 lMapPos;
		Vector3 lMapVecs[2];
		Vector3 vNormal;
		int size[2];
	};

	struct bsp_effect_t {
		char name[64];
		int brush;
		int unknown;
	};

	namespace {
		static char* extensions[] = { ".jpg", ".tga", NULL };
	}

	Loader_BSP::
	Loader_BSP(const Dir& dir,
		   const char* bspname) : m_dir(dir),
					  m_file(dir, (string("maps/")+bspname+".bsp").c_str()),
					  m_levelshot(new FileTexture(dir, (string("levelshots/")+bspname).c_str(), extensions)) {
		if (m_file.RWops() == NULL) {
			printf("Cannot open map %s\n", bspname);
			m_map = NULL;
		} else {
			printf("Reading map %s\n", bspname);
			m_map = new Map();
			read_bsp_file();
		}
	}

	void
	Loader_BSP::
	draw(const std::string& msg, unsigned int percent) {
		Font font;
		const Viewport& vp = Viewport::current();
		const int width = vp.w();
		const int height = vp.h();
		const int h_height = height >> 3;
		const int h = height>>6;
		const int w = width-20;
		const float fontsize = width*0.03;

		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		vp.begin_view_ortho();
		m_levelshot.bind();
		glBegin(GL_QUADS);
		glTexCoord2f(0, 0); glVertex2f(0, 0);
		glTexCoord2f(0, 1); glVertex2f(0, height);
		glTexCoord2f(1, 1); glVertex2f(width, height);
		glTexCoord2f(1, 0); glVertex2f(width, 0);
		glEnd();
		Texture::none().bind();
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0, 0, 0, 0.8);
		glBegin(GL_QUADS);
		glVertex2f(0, 0);
		glVertex2f(0, height>>2);
		glVertex2f(width, height>>2);
		glVertex2f(width, 0);
		glEnd();
		glColor4f(1, 1, 1, 1);
		glBegin(GL_QUADS);
		glVertex2f(10, h_height+5);
		glVertex2f(10+w*percent/100, h_height+5);
		glVertex2f(10+w*percent/100, h_height-5);
		glVertex2f(10, h_height-5);
		glEnd();
		glBegin(GL_LINE_STRIP);
		glVertex2f(8, h_height+7);
		glVertex2f(12+w, h_height+7);
		glVertex2f(12+w, h_height-7);
		glVertex2f(8, h_height-7);
		glVertex2f(8, h_height+7);
		glEnd();
		font.print(fontsize, h_height+h+7+fontsize, fontsize, "Loading BSP file...");
		font.print(fontsize, h_height-h-7, fontsize, msg.c_str());
		vp.end_view_ortho();

		SDL_GL_SwapBuffers();
	}

	void
	Loader_BSP::
	read_bsp_file() {
		bsp_header_t header;
		bsp_lumps_t  lumps;

		int		nb_vertices;
		int		nb_mesh_vertices;
		int*		mesh_vertices;
		Shader*		textures;
		Sky**		skies;
		Texture*	lightmaps;
		O3d**		meshes;
		int		i;

		m_file.read(&header, 1, sizeof(bsp_header_t));
		m_file.read(&lumps, 1, sizeof(bsp_lumps_t));

		// Read entities.
		/*
		{	char* entities;

			entities = new char[lumps.entities.length];

			m_file.seek(lumps.entities.offset, SEEK_SET);
			m_file.read(entities, 1, lumps.entities.length);

			Log << entities << "\n";

			delete [] entities;
		}
		*/

		draw("Loading vertices", 0);

		// Read vertices.
		nb_vertices = lumps.vertices.length / sizeof(VertexTLmNC);
		VertexTLmNC* vertices = new VertexTLmNC[nb_vertices];
		m_file.seek(lumps.vertices.offset, SEEK_SET);
		m_file.read(vertices, nb_vertices, sizeof(VertexTLmNC));
		for (i = 0; i < nb_vertices; ++i) {
			float tmp;

			tmp = vertices[i].vertex.y;
			vertices[i].vertex.y = vertices[i].vertex.z;
			vertices[i].vertex.z = -tmp;

			vertices[i].tex_coord.y = 1-vertices[i].tex_coord.y;
		}
		int va_contents =   VertexArray::vertices
				  | VertexArray::tex_coords
				  | VertexArray::lm_coords
				  | VertexArray::normals;
		VertexArray* va = new VertexArray(nb_vertices, va_contents);
		for (i = 0; i < nb_vertices; ++i) {
			va->vertex(i) = vertices[i].vertex;
			va->tex_coord(i) = vertices[i].tex_coord;
			va->lm_coord(i) = vertices[i].lightmap_coord;
			va->normal(i) = vertices[i].normal;
			//va->color(i) = vertices[i].color;
		}
		delete [] vertices;

		draw("Loading mesh vertices", 10);
		
		// Read mesh vertices.
		nb_mesh_vertices = lumps.mesh_vertices.length / sizeof(int);
		mesh_vertices = new int[nb_mesh_vertices];
		m_file.seek(lumps.mesh_vertices.offset, SEEK_SET);
		m_file.read(mesh_vertices, nb_mesh_vertices, sizeof(int));

		draw("Loading textures", 20);
		
		// Read textures.
		{	int nb_textures;
			bsp_texture_t tmp_texture;
			static char* extensions[] = { ".tga", ".jpg", NULL };

			nb_textures = lumps.textures.length / sizeof(bsp_texture_t);
			textures = new Shader[nb_textures];
			skies = new Sky*[nb_textures];

			m_file.seek(lumps.textures.offset, SEEK_SET);
			for (i = 0; i < nb_textures; i++) {
				m_file.read(&tmp_texture, 1, sizeof(bsp_texture_t));

				Q3Shader* q3shader = Q3ShaderManager.get_shader(tmp_texture.name);
				if (q3shader && q3shader->shader) {
					textures[i].set_shader(q3shader->shader);
					skies[i] = q3shader->sky;
				} else {
					Texture t(new FileTexture(m_dir, tmp_texture.name, extensions, true));

					textures[i].set_shader(Q3ShaderManager.get_shader("q3simple")->shader);
					textures[i].set_texture(t);
					skies[i] = NULL;
				}
			}
		}

		draw("Loading lightmaps", 30);

		int nb_lightmaps;
		// Read lightmaps.
		{	
			nb_lightmaps = lumps.lightmaps.length/49152;
			lightmaps = new Texture[nb_lightmaps];

			m_file.seek(lumps.lightmaps.offset, SEEK_SET);
			for (i = 0; i < nb_lightmaps; i++) {
				unsigned char* tmp_lightmap = new unsigned char[49152];

				m_file.read(tmp_lightmap, 1, 49152);
				lightmaps[i] = new LightMap(tmp_lightmap);
			}
		}

		// Read effects.
		{	int nb_effects;
			bsp_effect_t tmp_effect;

			nb_effects = lumps.effects.length / sizeof(bsp_effect_t);

			m_file.seek(lumps.effects.offset, SEEK_SET);

			for (i = 0; i < nb_effects; i++) {
				m_file.read(&tmp_effect, 1, sizeof(bsp_effect_t));
			}
		}

		draw("Loading faces", 40);

		int nb_faces;

		// Read faces
		{	
			bsp_face_t* faces;
			list<Topology*> topology_l;

			nb_faces = lumps.faces.length / sizeof(bsp_face_t);
			faces = new bsp_face_t[nb_faces];
			meshes = new O3d*[nb_faces];
			m_file.seek(lumps.faces.offset, SEEK_SET);
			m_file.read(faces, nb_faces, sizeof(bsp_face_t));

			for (i = 0; i < nb_faces; i++) {
				Mesh* mesh = NULL;

				if (skies[faces[i].texture] != NULL) {
					meshes[i] = skies[faces[i].texture];
					continue;
				}

				switch (faces[i].type) {
					case 1: {
						mesh = new Mesh;

						Topology& topology = *(new Topology(Topology::raw_triangle_fan, 2));

						topology[0] = faces[i].first_vertex;
						topology[1] = faces[i].nb_vertices;

						mesh->set_vertex_array(va);
						mesh->set_topology(&topology);
					} break;


					case 2: {
						mesh = new Mesh;

						vector<unsigned int> indices;
						Topology& topology = *(new Topology(Topology::grid, 3));
						int hsub = 8;
						int vsub = 8;
						VertexArray bpatch_src(*va, faces[i].first_vertex);

						{ // See if we could simplify the patch horizontally.
							int grid_size = faces[i].size[0]*faces[i].size[1];
							int j, k;
							bool aligned = true;

							for (k = 0; k < grid_size; k += faces[i].size[0]) {
								for (j = k; j < k+faces[i].size[0]-2; j += 2) {
									aligned = bpatch_src.aligned(j, j+1, j+2);
									if (!aligned) break;
								}
								if (!aligned) break;
							}

							if (aligned) {
								hsub = 1;
							}
						}

						{ // See if we could simplify the patch vertically.
							int grid_size = faces[i].size[0]*faces[i].size[1];
							int j, k;
							bool aligned = true;

							for (k = 0; k < grid_size-(faces[i].size[0]<<1); k += faces[i].size[0]<<1) {
								for (j = k; j < k+faces[i].size[0]; ++j) {
									aligned = bpatch_src.aligned(j, j+faces[i].size[0], j+(faces[i].size[0]<<1));
									if (!aligned) break;
								}
								if (!aligned) break;
							}

							if (aligned) {
								vsub = 1;
							}
						}

						topology[0] = ((faces[i].size[0]-1)>>1)*hsub+1;
						topology[1] = ((faces[i].size[1]-1)>>1)*vsub+1;
						topology[2] = 1;
						mesh->set_vertex_array(new VertexArray4BezierPatch(bpatch_src,
												  faces[i].size[0],
												  faces[i].size[1],
												  hsub, vsub));
						mesh->set_topology(&topology);
					} break;

					case 3: {
						mesh = new Mesh;

						Topology* topology = new Topology(Topology::triangles,
											      faces[i].nb_mesh_vertices);
						int j;
						list<Topology*>::iterator k;

						for (j = 0; j < faces[i].nb_mesh_vertices; j++) {
							(*topology)[j] = mesh_vertices[j+faces[i].first_mesh_vertex];
						}

						bool found = false;
						for (k = topology_l.begin(); k != topology_l.end(); ++k) {
							Topology* t = *k;

							if (*topology == *t) {
								delete topology;
								topology = t;
								found = true;
								break;
							}
						}
						if (!found) {
							topology_l.push_back(topology);
						}

						mesh->set_vertex_array(new VertexArray(*va, faces[i].first_vertex));
						mesh->set_topology(topology);
					} break;

					default:
						break;
				}

				if (mesh != NULL) {
					mesh->calculate_bounding_volume();
					
					{
						Shader shader(textures[faces[i].texture]);

						if (faces[i].lightmap >= 0) {
							shader.set_lightmap(lightmaps[faces[i].lightmap]);
						} else {
							shader.set_lightmap(Texture::none());
						}

						mesh->set_shader(shader);
					}

					m_map->m_mesh_l.push_back(mesh);
				}

				meshes[i] = mesh;
			}	

			delete [] faces;
			delete [] textures;
			delete [] lightmaps;
			delete [] mesh_vertices;
		}	

		// Read leafs and nodes.
		{
			BSP::Node**	leafs;
			BSP::Node**	nodes;
			int		nb_leafs;

			draw("Loading BSP leaves", 50);

			{
				bsp_leaf_t*	bsp_leafs;
				int*		bsp_leaf_faces;

				nb_leafs = lumps.leafs.length/sizeof(bsp_leaf_t);
				bsp_leafs = new bsp_leaf_t[nb_leafs];
				leafs = new BSP::Node*[nb_leafs];
				m_file.seek(lumps.leafs.offset, SEEK_SET);
				m_file.read(bsp_leafs, nb_leafs, sizeof(bsp_leaf_t));

				for (i = 0; i < nb_leafs; ++i) {
					BSP::Node* leaf;
					
					leafs[i] = leaf = new BSP::Node();
					Vector3 mins(bsp_leafs[i].mins[0],
						     bsp_leafs[i].mins[2],
						     -bsp_leafs[i].mins[1]);
					Vector3 maxs(bsp_leafs[i].maxs[0],
						     bsp_leafs[i].maxs[2],
						     -bsp_leafs[i].maxs[1]);

					leaf->set_bounding_volume(new BoundingBox(mins, maxs));
				}

				{
					int nb_leaf_faces = lumps.leaf_faces.length/sizeof(int);

					bsp_leaf_faces = new int[nb_leaf_faces];
					m_file.seek(lumps.leaf_faces.offset, SEEK_SET);
					m_file.read(bsp_leaf_faces, nb_leaf_faces, sizeof(int));
				}

				draw("Loading BSP leaf faces", 60);

				for (i = 0; i < nb_leafs; ++i) {
					BSP::Node* leaf = leafs[i];
					int j;

					for (j = 0; j < bsp_leafs[i].n_leaffaces; ++j) {
						int index = bsp_leaf_faces[bsp_leafs[i].leafface+j];

						if (index < 0 || index >= nb_faces) continue;

						O3d* mesh = meshes[index];

						if (mesh != NULL) {
							leaf->add_mesh(mesh);
						}
					}
				}

				draw("Loading PVS", 70);

				{
					int n_vecs;
					int sz_vecs;
					unsigned char* vecs;

					m_file.seek(lumps.visibility_data.offset, SEEK_SET);
					m_file.read(&n_vecs, 4, 1);
					m_file.read(&sz_vecs, 4, 1);
					vecs = new unsigned char[n_vecs*sz_vecs];
					m_file.read(vecs, sz_vecs, n_vecs);

					for (i = 0; i < nb_leafs; ++i) {
						int cluster_i = bsp_leafs[i].cluster;
						int j;

						if (cluster_i < 0 || cluster_i >= n_vecs) continue;

						for (j = 0; j < nb_leafs; ++j) {
							int cluster_j = bsp_leafs[j].cluster;

							if (cluster_j < 0 || cluster_j >= n_vecs) continue;

							if (vecs[cluster_j*sz_vecs+(cluster_i >> 3)] & (1 << (cluster_i & 7))) {
								leafs[i]->add_visible_leaf(leafs[j]);
							}
						}
					}

					delete [] vecs;
				}

				delete [] bsp_leafs;
				delete [] bsp_leaf_faces;
			}

			draw("Loading planes", 80);

			{
				int		nb_nodes;
				bsp_node_t*	bsp_nodes;
				Plane*		planes;

				nb_nodes = lumps.nodes.length/sizeof(bsp_node_t);
				bsp_nodes = new bsp_node_t[nb_nodes];
				nodes = new BSP::Node*[nb_nodes];
				m_file.seek(lumps.nodes.offset, SEEK_SET);
				m_file.read(bsp_nodes, nb_nodes, sizeof(bsp_node_t));

				// Read Planes.
				{
					int nb_planes = lumps.planes.length/sizeof(Plane);
					planes = new Plane[nb_planes];

					m_file.seek(lumps.planes.offset, SEEK_SET);
					m_file.read(planes, nb_planes, sizeof(Plane));
					for (i = 0; i < nb_planes; ++i) {
						float tmp = planes[i].b;
						planes[i].b = planes[i].c;
						planes[i].c = -tmp;
					}
				
				}

				for (i = 0; i < nb_nodes; ++i) {
					nodes[i] = new BSP::Node;
				}

				draw("Loading BSP nodes", 90);

				for (i = 0; i < nb_nodes; ++i) {
					BSP::Node* node = nodes[i];
					Vector3 mins(bsp_nodes[i].mins[0],
						     bsp_nodes[i].mins[2],
						     -bsp_nodes[i].mins[1]);
					Vector3 maxs(bsp_nodes[i].maxs[0],
						     bsp_nodes[i].maxs[2],
						     -bsp_nodes[i].maxs[1]);

					node->set_bounding_volume(new BoundingBox(mins, maxs));
					node->set_plane(new Plane(planes[bsp_nodes[i].plane]));
					if (bsp_nodes[i].children[0] >= 0) {
						int num_node = bsp_nodes[i].children[0];

						if (num_node < 0 || num_node >= nb_nodes) continue;
						node->set_front_node(nodes[num_node]);
					} else {
						int num_leaf = -1-bsp_nodes[i].children[0];

						if (num_leaf < 0 || num_leaf >= nb_leafs) continue;
						node->set_front_node(leafs[num_leaf]);
					}
					if (bsp_nodes[i].children[1] >= 0) {
						int num_node = bsp_nodes[i].children[1];

						if (num_node < 0 || num_node >= nb_nodes) continue;
						node->set_back_node(nodes[num_node]);
					} else {
						int num_leaf = -1-bsp_nodes[i].children[1];

						if (num_leaf < 0 || num_leaf >= nb_leafs) continue;
						node->set_back_node(leafs[num_leaf]);
					}
				}

				delete [] planes;
			}

			m_map->m_bsp_tree = nodes[0];

			draw("", 100);

		}

		delete [] meshes;
	}

	Map*
	Loader_BSP::
	read(const Dir& dir, const char* filename) {
		Loader_BSP loader(dir, filename);

		return loader.m_map;
	}

	Map*
	Loader_BSP::
	read(const char* bspname) {
		Loader_BSP loader(Dir("."), bspname);

		return loader.m_map;
	}

}

