/**************************************************************************

    This file is part of "The Science and Magick base library", a
    general library that provides miscellaneous features.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: lua.cpp,v 1.4 2003/08/24 15:45:00 rodzilla Exp $
*/

#ifndef NO_LUA

#include <signal.h>
#include <SaM3d/config_variable.h>
#include <SaM3d/log.h>
#include "lua.h"

// includes for the file lib
#include <stdio.h>			/* fileno */
#include <string.h>			/* strerror */
#include <stdlib.h>			/* fileno */
#include <errno.h>
#include <sys/stat.h>			/* stat, mkdir */
#include <sys/types.h>			/* dir, stat, mkdir */
#include <unistd.h>			/* getcwd, chdir, pathconf,  access */
#include <dirent.h>			/* getcwd, chdir, pathconf,  */
#include <fcntl.h>			/* for lock/unlock */

namespace SaM3d {

	using namespace std;

	namespace {

		int SaM3dLUA_file_dir (lua_State* L) {
			int n = lua_gettop(L);
			DIR *d = NULL;

			if (n == 0) {
				d = opendir(".");
			} else if (n == 1) {
				const char* dirname = luaL_checkstring(L, 1);

				if (dirname == NULL) {
					luaL_error(L, "optional argument to function 'dir' must be string");
					return 0;
				}
				d = opendir(dirname);
			} else {
				luaL_error(L, "function 'dir' takes at most one argument");
				return 0;
			}

			if (d == NULL) {
				lua_pushnil(L);
			} else {
				int i = 1;
				lua_newtable(L);
				struct dirent *entry;

				while ((entry = readdir(d)) != NULL) {
					lua_pushnumber(L, i++);
					lua_pushstring(L, entry->d_name);
					lua_rawset(L, -3);
				}
				closedir(d);
				lua_pushstring(L, "n");
				lua_pushnumber(L, --i);
				lua_rawset(L, -3);
			}

			return 1;
		}

		int SaM3dLUA_file_getcwd (lua_State* L) {
			const unsigned int LUA_GETCWD_STR_DIM=1024;
			char buf[LUA_GETCWD_STR_DIM];

			if (!getcwd(buf, LUA_GETCWD_STR_DIM)) {
				luaL_error(L, "error executing 'getcwd'");
				return 0;
			} else {
				lua_pushstring(L, buf);
			}

			return 1;
		}

		int SaM3dLUA_file_chdir (lua_State* L) {
			const char* path = luaL_checkstring(L, 1);

			if (chdir(path)) {
				luaL_error(L, "error executing 'chdir'");
				return 0;
			}

			lua_pushstring(L, path);
			return 1;
		}


		static struct luaL_reg filelib[] = {
			{"dir",		SaM3dLUA_file_dir},
			{"getcwd",	SaM3dLUA_file_getcwd},
			{"chdir",	SaM3dLUA_file_chdir},
			{NULL,		NULL}
		};

	}

	int lua_filelibopen(lua_State* L) {
		luaL_openlib(L, "file", filelib, 0);
		return 0;
	}

	static const luaL_reg lualibs[] = {
		{"baselib", lua_baselibopen},
		{"tablib", lua_tablibopen},
		{"iolib", lua_iolibopen},
		{"strlib", lua_strlibopen},
		{"mathlib", lua_mathlibopen},
		{"filelib", lua_filelibopen},
		{"dblib", lua_dblibopen},
		{NULL, NULL}
	};

	namespace {
		int SaM3dLUA_print (lua_State *L) {
			int n = lua_gettop(L);  /* number of arguments */
			int i;
			lua_getglobal(L, "tostring");
			for (i=1; i<=n; i++) {
				const char *s;
				lua_pushvalue(L, -1);  /* function to be called */
				lua_pushvalue(L, i);   /* value to print */
				lua_call(L, 1, 1);
				s = lua_tostring(L, -1);  /* get result */
				if (s == NULL) {
					return luaL_error(L, "`tostring' must return a string to `print'");
				}
				Log << s;
				lua_pop(L, 1);  /* pop result */
			}
			Log << "\n";
			return 0;
		}

		int SaM3dLUA_index (lua_State *L) {
			const char* var_name = lua_tostring(L, -1);
			CfgVar* cfg_var = CfgVar::get(var_name);

			if (cfg_var == NULL) {
				lua_rawget(L, -2);
			} else {
				bool bval;
				float fval;

				if (cfg_var->get(bval)) {
					lua_pushboolean(L, bval);
				} else if (cfg_var->get(fval)) {
					lua_pushnumber(L, fval);
				} else {
					lua_pushnil(L);
				}
			}

			return 1;
		}

		int SaM3dLUA_newindex (lua_State *L) {
			const char* var_name = lua_tostring(L, -2);
			CfgVar* cfg_var = CfgVar::get(var_name);

			if (cfg_var == NULL) {
				lua_rawset(L, -3);
			} else {
				if (lua_isboolean(L, -1)) {
					bool value = lua_toboolean(L, -1);
					cfg_var->set(value);
				} else if (lua_isnumber(L, -1)) {
					float value = lua_tonumber(L, -1);
					cfg_var->set(value);
				} else if (lua_isstring(L, -1)) {
					string value = lua_tostring(L, -1);
					cfg_var->set(value);
				} else {
					Log << "Bad type for configuration variable " << var_name << ".\n";
				}
			}

			return 0;
		}

	}

	CLua::
	CLua() {
		const luaL_reg *lib = lualibs;

		m_lua_state = lua_open();
		for (; lib->name; lib++) {
			Log << "Loading lua library \"" << lib->name << "\"\n";
			lib->func(m_lua_state);
			lua_settop(m_lua_state, 0);
		}

		register_cfunction("print", SaM3dLUA_print);
		
		lua_newtable(m_lua_state);
		lua_pushstring(m_lua_state, "__index");
		lua_pushcfunction(m_lua_state, SaM3dLUA_index);
		lua_rawset(m_lua_state, -3);
		lua_pushstring(m_lua_state, "__newindex");
		lua_pushcfunction(m_lua_state, SaM3dLUA_newindex);
		lua_rawset(m_lua_state, -3);
		lua_setmetatable(m_lua_state, LUA_GLOBALSINDEX);

		m_lua_command = "";

		Console.set_interpreter(this);
	}

	CLua::
	~CLua() {
		lua_close(m_lua_state);

		Console.set_interpreter(NULL);
	}

	CLua&
	CLua::
	instance() {
		static CLua instance;

		return instance;
	}

	void
	CLua::
	register_cfunction(const char* n, lua_CFunction f) {
		lua_register(m_lua_state, n, f);
	}

	void
	CLua::
	openlib(const char *libname, const luaL_reg *l) {
		luaL_openlib(m_lua_state, libname, l, 1);
	}

	void
	CLua::
	do_file(const char *filename) {
		lua_dofile(m_lua_state, filename);
	}

	void
	CLua::
	do_string(const string& s) {
		int status;
		int stacktop = lua_gettop(m_lua_state);
		string lua_command = m_lua_command + s;

		m_lua_command = "";
		lua_pushstring(m_lua_state, lua_command.c_str());
		status = luaL_loadbuffer(m_lua_state, lua_tostring(m_lua_state, stacktop+1),
					 lua_strlen(m_lua_state, stacktop+1), "=stdin");
		if (   (status == LUA_ERRSYNTAX)
		    && (strstr(lua_tostring(m_lua_state, -1), "near `<eof>'") != NULL)) {
			m_lua_command = lua_command;
		} else {
			lua_remove(m_lua_state, stacktop+1);
			if (status == 0) {
				int base = lua_gettop(m_lua_state);
				lua_pushliteral(m_lua_state, "_TRACEBACK");
				lua_rawget(m_lua_state, LUA_GLOBALSINDEX);
				lua_insert(m_lua_state, base); 
				status = lua_pcall(m_lua_state, 0, LUA_MULTRET, base);
				lua_remove(m_lua_state, base);
			}
			if (status != 0) {
				const char* msg = lua_tostring(m_lua_state, -1);
				if (msg == NULL) {
					msg = "(error with no message)";
				}
				Log.printf("Error: %s\n", msg);
				lua_pop(m_lua_state, 1);
			}
			/*
			if (status == 0 && lua_gettop(m_lua_state) > stacktop) {
				lua_getglobal(m_lua_state, "print");
				lua_insert(m_lua_state, 1);
				if (lua_pcall(m_lua_state, lua_gettop(m_lua_state)-stacktop-1, 0, 0) != 0) {
					Log.printf("Error calling `print'\n");
				}
			}
			*/
		}
		lua_settop(m_lua_state, stacktop);
	}

}

#endif
