/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: mesh.cpp,v 1.6 2003/08/31 10:04:34 rodzilla Exp $

    The Mesh class handle meshes.
*/

#include <assert.h>
#include <unistd.h>
#include <iostream>
#include <SaM3d/opengl.h>
#include <SaM3d/performance_counter.h>
#include "mesh.h"

namespace SaM3d {

	using namespace std;

	Mesh::
	Mesh() {
		m_vertex_array = NULL;
		m_topology = NULL;
		m_bounding_volume = NULL;
	}

	Mesh::
	~Mesh() {
		if (m_vertex_array) {
		}
		if (m_topology) {
		}
	}

	void
	Mesh::
	set_topology(Topology* t) {
		assert(t != NULL);
		if (m_topology != t) {
			m_topology = t;
		}
	}

	void
	Mesh::
	set_vertex_array(VertexArray* a) {
		assert(a != NULL);
		if (m_vertex_array != a) {
			m_vertex_array = a;
		}
	}

	void
	Mesh::
	set_shader(const Shader& shader) {
		m_shader = shader;
	}

	void
	Mesh::
	draw() {
		m_shader.draw(m_vertex_array, m_topology);
		/*
		m_texture.bind();
		Texture::set_texture_unit(1);
		m_lightmap.bind();
		Texture::set_texture_unit(0);

		m_vertex_array->glPointer();
		m_topology->draw();
		*/
	}

	void
	Mesh::
	calculate_bounding_volume() {
		delete m_bounding_volume;
		if (m_topology) {
			m_bounding_volume = m_topology->calculate_bounding_volume(m_vertex_array);
		} else {
			m_bounding_volume = NULL;
		}
	}


	Topology::
	Topology() {
	}

	Topology::
	Topology(const primitive_t primitive,
	         unsigned int nb_indices,
		 unsigned int* indices) : m_primitive(primitive),
		 			    m_nb_indices(nb_indices) {
		if (indices) {
			m_indices = indices;
			m_own_indices = false;
		} else {
			m_indices = new unsigned int[m_nb_indices];
			m_own_indices = true;
		}
	}

	Topology::
	~Topology() {
		delete [] m_indices;
	}

	unsigned int&
	Topology::
	operator [](const unsigned int index) {
		return m_indices[index];
	}

	bool
	Topology::
	operator ==(const Topology& that) const {
		unsigned int i;

		if (m_primitive != that.m_primitive) return false;
		if (m_nb_indices != that.m_nb_indices) return false;
		for (i = 0; i < m_nb_indices; ++i) {
			if (m_indices[i] != that.m_indices[i]) {
				return false;
			}
		}
		return true;
	}

	void
	Topology::
	draw() {
		unsigned int i, j;

		switch (m_primitive) {
			case triangles:
				glDrawElements(GL_TRIANGLES, m_nb_indices, GL_UNSIGNED_INT, m_indices);
				PerformanceCounter.add_triangles(m_nb_indices/3);
				break;

			case raw_triangle_fan:
				for (i = 0; i < m_nb_indices; i += 2) {
					glDrawArrays(GL_TRIANGLE_FAN, m_indices[i], m_indices[i+1]);
					PerformanceCounter.add_triangles(m_indices[i+1]-2);
				}
				break;

			case grid:
				for (i = m_indices[2]; i < m_indices[1]; i += m_indices[2]) {
					static unsigned int elements[1024];
					unsigned int k = 0;

					for (j = 0; j < m_indices[0]; j += m_indices[2]) {
						elements[k++] = (i-m_indices[2])*m_indices[0]+j;
						elements[k++] = i*m_indices[0]+j;
					}

					glDrawElements(GL_TRIANGLE_STRIP, k, GL_UNSIGNED_INT, elements);
					PerformanceCounter.add_triangles((m_indices[0]-1)*(m_indices[1]-1)*2/m_indices[2]/m_indices[2]);
				}
				break;

			default:
				break;
		}
	}

	void refine_mins_maxs(Vector3& mins, Vector3& maxs, const Vector3& tmp) {
		if (tmp.x < mins.x) mins.x = tmp.x;
		if (tmp.y < mins.y) mins.y = tmp.y;
		if (tmp.z < mins.z) mins.z = tmp.z;
		if (tmp.x > maxs.x) maxs.x = tmp.x;
		if (tmp.y > maxs.y) maxs.y = tmp.y;
		if (tmp.z > maxs.z) maxs.z = tmp.z;
	}

	BoundingVolume*
	Topology::
	calculate_bounding_volume(VertexArray* va) {
		if (va == NULL) return NULL;

		switch (m_primitive) {
			case triangles: {
				unsigned i;
				Vector3 mins = va->vertex(0);
				Vector3 maxs = va->vertex(0);

				for (i = 0; i < m_nb_indices; ++i) {
					refine_mins_maxs(mins, maxs, va->vertex(m_indices[i]));
				}

				return new BoundingBox(mins,maxs);
			}	break;

			case raw_triangle_fan: {
				unsigned i, j;
				Vector3 mins = va->vertex(0);
				Vector3 maxs = va->vertex(0);

				for (i = 0; i < m_nb_indices; i += 2) {
					for (j = 0; j < m_indices[i+1]; ++j) {
						refine_mins_maxs(mins, maxs, va->vertex(m_indices[i]+j));
					}
				}

				return new BoundingBox(mins,maxs);
			}	break;

			case grid: {
				unsigned i, j;
				Vector3 mins = va->vertex(0);
				Vector3 maxs = va->vertex(0);

				for (i = 0; i < m_indices[1]; i += m_indices[2]) {
					for (j = 0; j < m_indices[0]; j += m_indices[2]) {
						refine_mins_maxs(mins, maxs, va->vertex(i*m_indices[0]+j));
					}
				}

				return new BoundingBox(mins,maxs);
			}	break;

			default:
				break;
		}

		return NULL;
	}

}

