/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2003 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: q3parser.cpp,v 1.2 2003/08/31 10:04:34 rodzilla Exp $
*/

#include <SaM3d/file_texture.h>
#include <SaM3d/log.h>
#include "animated_texture.h"
#include "q3parser.h"

using namespace std;
using namespace SaM3d;

CQ3ShaderManager::Parser::
Parser(const Dir& d, const string& filename) : m_dir(d),
					       m_filename(filename),
					       m_lex(d, filename) {
}

Waveform*
CQ3ShaderManager::Parser::
parse_waveform() {
	string waveform;
	float base, amplitude, phase, freq;

	if (   m_lex.get_identifier(waveform)
	    && m_lex.get_float(base)
	    && m_lex.get_float(amplitude)
	    && m_lex.get_float(phase)
	    && m_lex.get_float(freq)) {
		if (!strcasecmp(waveform.c_str(), "sin")) {
			return new Waveform(Waveform::sine, base, amplitude, phase, freq);
		} else if (!strcasecmp(waveform.c_str(), "triangle")) {
			return new Waveform(Waveform::triangle, base, amplitude, phase, freq);
		} else if (!strcasecmp(waveform.c_str(), "square")) {
			return new Waveform(Waveform::square, base, amplitude, phase, freq);
		} else if (!strcasecmp(waveform.c_str(), "sawtooth")) {
			return new Waveform(Waveform::sawtooth, base, amplitude, phase, freq);
		} else if (!strcasecmp(waveform.c_str(), "inversesawtooth")) {
			return new Waveform(Waveform::inverse_sawtooth, base, amplitude, phase, freq);
		} else if (!strcasecmp(waveform.c_str(), "noise")) {
			return new Waveform(Waveform::inverse_sawtooth, base, amplitude, phase, freq);
		}
	}
	return NULL;
}

GLenum blendfunc_from_name(const char* name) {
	struct blendfunc {
		const char* name;
		GLenum value;
	};
	static blendfunc names[] = {
		{ "GL_DST_COLOR", GL_DST_COLOR },
		{ "GL_ONE ", GL_ONE  },
		{ "GL_ONE", GL_ONE },
		{ "GL_ONE_MINUS_DST_COLOR", GL_ONE_MINUS_DST_COLOR },
		{ "GL_ONE_MINUS_SRC_ALPHA", GL_ONE_MINUS_SRC_ALPHA },
		{ "GL_ONE_MINUS_SRC_COLOR", GL_ONE_MINUS_SRC_COLOR },
		{ "GL_SRC_ALPHA", GL_SRC_ALPHA },
		{ "GL_SRC_COLOR", GL_SRC_COLOR },
		{ "GL_ZERO", GL_ZERO },
		{ NULL, 0 }
	};
	int i;

	for (i = 0; names[i].name != NULL; ++i) {
		if (!strcasecmp(name, names[i].name)) {
			return names[i].value;
		}
	}

	return GL_ZERO;
}

void
CQ3ShaderManager::Parser::
parse() {
	Log << "Parsing shader file \"" << m_filename << "\".\n";

	for (;;) {
		string shadername;

		m_lex.skip_spaces_and_endl();
		if (!m_lex.get_path(shadername)) {
			break;
		}
		//Log << "Reading shader " << shadername << ".\n";
		Q3Shader* q3shader = new Q3Shader;
		Q3ShaderManager.m_shaders[shadername] = q3shader;

		ComplexShader* shader = new ComplexShader;
		q3shader->shader = shader;
		q3shader->sky = NULL;

		bool mipmap = true;

		m_lex.skip_spaces_and_endl();
		if (!m_lex.get_special('{')) {
			Log << "syntax error (1) in file \"" << m_filename << "\".\n";
			return;
		}

		for (;;) {
			string keyword;
			
			m_lex.skip_spaces_and_endl();
			if (!m_lex.get_identifier(keyword)) {
				if (m_lex.get_special('}')) {
					break;
				} else if (!m_lex.get_special('{')) {
					Log << "syntax error (2) in file \"" << m_filename << "\".\n";
					return;
				} else {
					ComplexShader::Layer* layer = new ComplexShader::Layer;
					bool layer_blend = false;
					bool syntax_error = true;

					for(;;) {
						m_lex.skip_spaces_and_endl();
						if (!m_lex.get_identifier(keyword)) {
							if (m_lex.get_special('}')) {
								syntax_error = false;
								if (   shader->add_layer(layer)
								    && layer_blend) {
									shader->m_blend = true;
								}
							}
							break;
						}
						//Log << keyword << "\n";
						if (!strcasecmp(keyword.c_str(), "alphafunc")) {
							string tmp;

							if (!m_lex.get_identifier(tmp)) {
								break;
							} else if (!strcasecmp(tmp.c_str(), "gt0")) {
								layer->m_alphafunc = GL_GREATER;
								layer->m_alpharef = 0.0;
							} else if (!strcasecmp(tmp.c_str(), "lt128")) {
								layer->m_alphafunc = GL_LESS;
								layer->m_alpharef = 0.5;
							} else if (!strcasecmp(tmp.c_str(), "ge128")) {
								layer->m_alphafunc = GL_GEQUAL;
								layer->m_alpharef = 0.5;
							} else {
								break;
							}
						} else if (!strcasecmp(keyword.c_str(), "alphagen")) {
							string tmp;

							if (!m_lex.get_identifier(tmp)) {
								break;
							} else if (!strcasecmp(tmp.c_str(), "const")) {
								float f;

								if (!m_lex.get_float(f)) {
									break;
								}
							} else if (!strcasecmp(tmp.c_str(), "entity")) {
							} else if (!strcasecmp(tmp.c_str(), "lightingspecular")) {
							} else if (!strcasecmp(tmp.c_str(), "portal")) {
								float f;

								if (!m_lex.get_float(f)) {
									break;
								}
							} else if (!strcasecmp(tmp.c_str(), "vertex")) {
							} else if (!strcasecmp(tmp.c_str(), "wave")) {
								Waveform* waveform = parse_waveform();
								if (waveform == NULL) {
									break;
								}
								layer->m_alphagen = waveform;
							} else {
								break;
							}
						} else if (!strcasecmp(keyword.c_str(), "alphamap")) {
							string tmp;

							if (m_lex.get_path(tmp)) {
								continue;
							}
						} else if (!strcasecmp(keyword.c_str(), "animmap")) {
							float f;
							Texture textures[16];
							unsigned int i;

							if (!m_lex.get_float(f)) {
								break;
							}

							for (i = 0;;++i) {
								string texture_name;

								if (!m_lex.get_path(texture_name)) {
									break;
								}

								string prefix(texture_name, 0, texture_name.size()-4);
								string suffix(texture_name, texture_name.size()-4, 4);

								if (suffix == ".tga") {
									static char* extensions[] = { ".tga", ".jpg", NULL };

									textures[i] = new FileTexture(m_dir, prefix.c_str(), extensions, false);
								} else if (suffix == ".jpg") {
									static char* extensions[] = { ".jpg", ".tga", NULL };

									textures[i] = new FileTexture(m_dir, prefix.c_str(), extensions, false);
								} else {
									static char* extensions[] = { "", ".tga", ".jpg", NULL };

									textures[i] = new FileTexture(m_dir, texture_name.c_str(), extensions, false);
								}
							}

							layer->m_texture = ComplexShader::Layer::layertexture;
							layer->m_layertexture = new AnimatedTexture(f, textures, i);

						} else if (!strcasecmp(keyword.c_str(), "blendfunc")) {
							string tmp1, tmp2;

							if (!m_lex.get_identifier(tmp1)) {
								break;
							} else if (!strcasecmp(tmp1.c_str(), "add")) {
								layer->m_sfactor = GL_ONE;
								layer->m_dfactor = GL_ONE;
							} else if (!strcasecmp(tmp1.c_str(), "blend")) {
								layer->m_sfactor = GL_SRC_ALPHA;
								layer->m_dfactor = GL_ONE_MINUS_SRC_ALPHA;
							} else if (!strcasecmp(tmp1.c_str(), "filter")) {
								layer->m_sfactor = GL_ZERO;
								layer->m_dfactor = GL_SRC_COLOR;
							} else {
								m_lex.get_identifier(tmp2);
								layer->m_sfactor = blendfunc_from_name(tmp1.c_str());
								layer->m_dfactor = blendfunc_from_name(tmp2.c_str());
							}

							layer_blend = true;
						} else if (!strcasecmp(keyword.c_str(), "clampmap")) {
							string texture_name;

							if (!m_lex.get_path(texture_name)) {
								break;
							}

							if (texture_name == "$lightmap") {
								layer->m_texture = ComplexShader::Layer::lightmap;
								layer->m_tcgen = ComplexShader::Layer::tc_lightmap;
							} else {
								string prefix(texture_name, 0, texture_name.size()-4);
								string suffix(texture_name, texture_name.size()-4, 4);

								layer->m_texture = ComplexShader::Layer::layertexture;
								if (suffix == ".tga") {
									static char* extensions[] = { ".tga", ".jpg", NULL };

									layer->m_layertexture = new FileTexture(m_dir, prefix.c_str(), extensions, mipmap, GL_CLAMP);
								} else if (suffix == ".jpg") {
									static char* extensions[] = { ".jpg", ".tga", NULL };

									layer->m_layertexture = new FileTexture(m_dir, prefix.c_str(), extensions, mipmap, GL_CLAMP);
								} else {
									static char* extensions[] = { "", ".tga", ".jpg", NULL };

									layer->m_layertexture = new FileTexture(m_dir, texture_name.c_str(), extensions, mipmap, GL_CLAMP);
								}
							}
						} else if (!strcasecmp(keyword.c_str(), "depthfunc")) {
							string tmp;

							if (!m_lex.get_path(tmp)) {
								break;
							}
						} else if (!strcasecmp(keyword.c_str(), "detail")) {
						} else if (!strcasecmp(keyword.c_str(), "depthwrite")) {
						} else if (!strcasecmp(keyword.c_str(), "fog")) {
							string tmp;

							if (!m_lex.get_identifier(tmp)) {
								break;
							}
						} else if (   !strcasecmp(keyword.c_str(), "map")
						           || !strcasecmp(keyword.c_str(), "mapcomp")
						           || !strcasecmp(keyword.c_str(), "mapnocomp")) {
							string texture_name;

							if (!m_lex.get_path(texture_name)) {
								break;
							}

							if (texture_name == "$lightmap") {
								layer->m_texture = ComplexShader::Layer::lightmap;
								layer->m_tcgen = ComplexShader::Layer::tc_lightmap;
							} else {
								string prefix(texture_name, 0, texture_name.size()-4);
								string suffix(texture_name, texture_name.size()-4, 4);

								layer->m_texture = ComplexShader::Layer::layertexture;
								if (suffix == ".tga") {
									static char* extensions[] = { ".tga", ".jpg", NULL };

									layer->m_layertexture = new FileTexture(m_dir, prefix.c_str(), extensions, mipmap);
								} else if (suffix == ".jpg") {
									static char* extensions[] = { ".jpg", ".tga", NULL };

									layer->m_layertexture = new FileTexture(m_dir, prefix.c_str(), extensions, mipmap);
								} else {
									static char* extensions[] = { "", ".tga", ".jpg", NULL };

									layer->m_layertexture = new FileTexture(m_dir, texture_name.c_str(), extensions, mipmap);
								}
							}
						} else if (!strcasecmp(keyword.c_str(), "rgbgen")) {
							string tmp;

							if (!m_lex.get_identifier(tmp)) {
								break;
							} else if (!strcasecmp(tmp.c_str(), "const")) {
								float f1, f2, f3;

				 				if (!(   m_lex.get_special('(')
								      && m_lex.get_float(f1)
								      && m_lex.get_float(f2)
								      && m_lex.get_float(f3)
								      && m_lex.get_special(')'))) {
									break;
								}
							} else if (!strcasecmp(tmp.c_str(), "diffuse")) {
							} else if (!strcasecmp(tmp.c_str(), "entity")) {
							} else if (!strcasecmp(tmp.c_str(), "exactvertex")) {
							} else if (!strcasecmp(tmp.c_str(), "identity")) {
							} else if (!strcasecmp(tmp.c_str(), "identitylighting")) {
							} else if (!strcasecmp(tmp.c_str(), "lightingdiffuse")) {
							} else if (!strcasecmp(tmp.c_str(), "vertex")) {
							} else if (!strcasecmp(tmp.c_str(), "wave")) {
								Waveform* waveform = parse_waveform();
								if (waveform == NULL) {
									break;
								}
								layer->m_rgbgen = waveform;
							} else {
								break;
							}
						} else if (!strcasecmp(keyword.c_str(), "tcgen")) {
							string tmp;

							if (!m_lex.get_identifier(tmp)) {
								break;
							} else if (!strcasecmp(tmp.c_str(), "environment")) {
								layer->m_tcgen = ComplexShader::Layer::tc_environment;
							} else if (!strcasecmp(tmp.c_str(), "vector")) {
								float f1, f2, f3, f4, f5, f6;

				 				if (!(   m_lex.get_special('(')
								      && m_lex.get_float(f1)
								      && m_lex.get_float(f2)
								      && m_lex.get_float(f3)
								      && m_lex.get_special(')')
								      && m_lex.get_special('(')
								      && m_lex.get_float(f4)
								      && m_lex.get_float(f5)
								      && m_lex.get_float(f6)
								      && m_lex.get_special(')'))) {
									break;
								}
							} else {
								break;
							}
						} else if (!strcasecmp(keyword.c_str(), "tcmod")) {
							string tmp;
							float f1, f2, f3, f4, f5, f6;

							if (!m_lex.get_identifier(tmp)) {
								break;
							} else if (!strcasecmp(tmp.c_str(), "rotate")) {
								if (!m_lex.get_float(f1)) {
									break;
								}

								layer->add_tcmod(new TCMod_rotate(f1));

								float foo;
								// bug in textures/sfx/portal_sfx ???
								m_lex.get_float(foo);
							} else if (!strcasecmp(tmp.c_str(), "scale")) {
								if (!(   m_lex.get_float(f1)
								      && m_lex.get_float(f2))) {
									break;
								}

								layer->add_tcmod(new TCMod_scale(f1, f2));

								float foo;
								// bug in textures/eerie/iron01_d_flicker ???
								m_lex.get_float(foo);
								m_lex.get_float(foo);
							} else if (!strcasecmp(tmp.c_str(), "scroll")) {
								if (!(   m_lex.get_float(f1)
								      && m_lex.get_float(f2))) {
									break;
								}

								layer->add_tcmod(new TCMod_translate(f1, f2));

								float foo;
								// bug in textures/eerie/iron01_d_flicker ???
								m_lex.get_float(foo);
								m_lex.get_float(foo);
							} else if (!strcasecmp(tmp.c_str(), "stretch")) {
								Waveform* waveform = parse_waveform();
								if (!waveform) {
									break;
								}

								layer->add_tcmod(new TCMod_stretch(waveform));

							} else if (!strcasecmp(tmp.c_str(), "transform")) {
								if (!(   m_lex.get_float(f1)
								      && m_lex.get_float(f2)
								      && m_lex.get_float(f3)
								      && m_lex.get_float(f4)
								      && m_lex.get_float(f5)
								      && m_lex.get_float(f6))) {
									break;
								}
							} else if (!strcasecmp(tmp.c_str(), "turb")) {
								if (!(   (   m_lex.get_float(f1)
									  && m_lex.get_float(f2)
									  && m_lex.get_float(f3)
									  && m_lex.get_float(f4))
								      || parse_waveform())) {
									break;
								}

								layer->add_tcmod(new TCMod_turb(f2, f3, f4));
							}
						} else {
							Log << "Unknown keyword " << keyword << ".\n";
							break;
						} 
					}
					if (syntax_error) {
						delete layer;
						Log << "syntax error (2) in file \"" << m_filename << "\".\n";
						return;
					}
				}
				continue;
			}
			//Log << keyword << "\n";
			if (!strcasecmp(keyword.c_str(), "allowcompress")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "backsided")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "cloudparms")) {
				string tmp;
				float f;

				if (m_lex.get_float(f)) {
					m_lex.get_identifier(tmp);
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "cull")) {
				string tmp;

				if (m_lex.get_identifier(tmp)) {
					if (   !strcasecmp(tmp.c_str(), "disable")
					    || !strcasecmp(tmp.c_str(), "twosided")
					    || !strcasecmp(tmp.c_str(), "none")) {
						shader->m_cullface = GL_NONE;
					} else if (!strcasecmp(tmp.c_str(), "back")) {
						shader->m_cullface = GL_BACK;
					}
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "deformvertexes")) {
				string tmp;
				float f1, f2, f3;

				if (!m_lex.get_identifier(tmp)) {
				} else if (!strcasecmp(tmp.c_str(), "autosprite")) {
					continue;
				} else if (!strcasecmp(tmp.c_str(), "autosprite2")) {
					continue;
				} else if (!strcasecmp(tmp.c_str(), "bulge")) {
					if (   m_lex.get_float(f1)
					    && m_lex.get_float(f2)
					    && m_lex.get_float(f3)) {
						continue;
					}
				} else if (!strcasecmp(tmp.c_str(), "move")) {
					if (   m_lex.get_float(f1)
					    && m_lex.get_float(f2)
					    && m_lex.get_float(f3)
					    && parse_waveform()) {
						continue;
					}
				} else if (!strcasecmp(tmp.c_str(), "normal")) {
					if (   m_lex.get_float(f1)
					    && m_lex.get_float(f2)) {
						continue;
					}
				} else if (!strcasecmp(tmp.c_str(), "projectionshadow")) {
					continue;
				} else if (!strcasecmp(tmp.c_str(), "text0")) {
					continue;
				} else if (!strcasecmp(tmp.c_str(), "text1")) {
					continue;
				} else if (!strcasecmp(tmp.c_str(), "wave")) {
					if (   m_lex.get_float(f1)
					    && parse_waveform()) {
						continue;
					}
				}
			} else if (!strcasecmp(keyword.c_str(), "entitymergable")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "foggen")) {
				if (parse_waveform()) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "fogonly")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "fogparms")) {
				float f1, f2, f3, f4, f5;

				if (m_lex.get_special('(')) {
					if (   m_lex.get_float(f1)
					    && m_lex.get_float(f2)
					    && m_lex.get_float(f3)	
					    && m_lex.get_special(')')	
					    && m_lex.get_float(f4)) {
						continue;
					}
				} else if (   m_lex.get_float(f1)
					   && m_lex.get_float(f2)
					   && m_lex.get_float(f3)
					   && m_lex.get_float(f4)
					   && m_lex.get_float(f5)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "fogvars")) {
				float f1, f2, f3, f4;

				if (m_lex.get_special('(')
				    && m_lex.get_float(f1)
				    && m_lex.get_float(f2)
				    && m_lex.get_float(f3)	
				    && m_lex.get_special(')')	
				    && m_lex.get_float(f4)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "light")) {
				float f;

				if (m_lex.get_float(f)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "lightgridmulamb")) {
				float f;

				if (m_lex.get_float(f)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "light1")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "lightning")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "nocompress")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "nofog")) {
				continue;
			} else if (   !strcasecmp(keyword.c_str(), "nomipmap")
				   || !strcasecmp(keyword.c_str(), "nomipmaps")) {
				mipmap = false;
				continue;
			} else if (!strcasecmp(keyword.c_str(), "nopicmip")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "picmip2")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "polygonoffset")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "portal")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "q3map_globaltexture")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "q3map_backsplash")) {
				float f1, f2;

				if (   m_lex.get_float(f1)
				    && m_lex.get_float(f2)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "q3map_flare")) {
				string tmp;

				if (m_lex.get_identifier(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "q3map_lightimage")) {
				string tmp;

				if (m_lex.get_path(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "q3map_lightsubdivide")) {
				float tmp;

				if (m_lex.get_float(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "q3map_sun")) {
				float f1, f2, f3, f4, f5, f6;

				if (   m_lex.get_float(f1)
				    && m_lex.get_float(f2)
				    && m_lex.get_float(f3)
				    && m_lex.get_float(f4)
				    && m_lex.get_float(f5)
				    && m_lex.get_float(f6)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "q3map_surfacelight")) {
				float tmp;

				if (m_lex.get_float(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "qer_editorimage")) {
				string tmp;

				if (m_lex.get_path(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "qer_lightimage")) {
				string tmp;

				if (m_lex.get_path(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "qer_nocarve")) {
				continue;
			} else if (!strcasecmp(keyword.c_str(), "qer_trans")) {
				float tmp;

				if (m_lex.get_float(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "sky")) {
				string far_prefix;

				if (m_lex.get_path(far_prefix)) {
					q3shader->sky = new Sky(new SkyBox(m_dir, far_prefix), NULL, NULL);
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "skyfogvars")) {
				float f1, f2, f3, f4;

				if (m_lex.get_special('(')
				    && m_lex.get_float(f1)
				    && m_lex.get_float(f2)
				    && m_lex.get_float(f3)	
				    && m_lex.get_special(')')	
				    && m_lex.get_float(f4)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "skyparms")) {
				string far_prefix, near_prefix, tmp;
				float cloudheight = 0;

				if (   m_lex.get_path(far_prefix)
				    && (   m_lex.get_special('-')
					|| m_lex.get_float(cloudheight)
					|| m_lex.get_identifier(tmp))
				    && m_lex.get_path(near_prefix)) {
					SkyBox* far = NULL;
					SkyDome* clouds = NULL;
					SkyBox* near = NULL;

					if (   far_prefix != "-"
					    && far_prefix != "full") {
						far = new SkyBox(m_dir, far_prefix);
					}
					if (cloudheight != 0) {
						clouds = new SkyDome(Shader(shader, Texture::none(), Texture::none()), cloudheight);
					}
					if (   near_prefix != "-"
					    && near_prefix != "full") {
						near = new SkyBox(m_dir, near_prefix);
					}

					q3shader->sky = new Sky(far, clouds, near);
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "sort")) {
				string tmp;

				if (m_lex.get_identifier(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "surfaceparm")) {
				string tmp;

				if (m_lex.get_identifier(tmp)) {
					if (!strcasecmp(tmp.c_str(), "trans")) {
						shader->m_blend = true;
					}
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "tesssize")) {
				float tmp;

				if (m_lex.get_float(tmp)) {
					continue;
				}
			} else if (!strcasecmp(keyword.c_str(), "waterfogvars")) {
				float f1, f2, f3, f4;

				if (m_lex.get_special('(')
				    && m_lex.get_float(f1)
				    && m_lex.get_float(f2)
				    && m_lex.get_float(f3)	
				    && m_lex.get_special(')')	
				    && m_lex.get_float(f4)) {
					continue;
				}
			} else {
				Log << "Unknown keyword " << keyword << ".\n";
				return;
			}
			Log << "syntax error (3) in file \"" << m_filename << "\".\n";
			return;
		}
	}
}

