/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2002 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: shader.cpp,v 1.6 2003/08/31 10:04:34 rodzilla Exp $
*/

#include <math.h>
#include <SaM3d/log.h>
#include <SaM3d/loop.h>
#include "mesh.h"
#include "shader.h"
#include "vertex_array.h"

namespace SaM3d {

	using namespace std;

	
	Waveform::
	Waveform(const waveform_t waveform,
		 const float base,
		 const float amplitude,
		 const float phase,
		 const float frequency) : m_waveform(waveform),
					  m_base(base),
					  m_amplitude(amplitude),
					  m_phase(phase),
					  m_frequency(frequency) {
	}

	float
	Waveform::
	value(const float x) {
		float fx = (x+m_phase)*m_frequency;
		float rv;

		fx -= floorf(fx);

		switch (m_waveform) {
			case sine:
				rv = sin(fx*2.0*M_PI);
				break;
			case triangle:
				rv = (fx < 0.5) ? 4*fx-1 : 3-4*fx;
				break;
			case square:
				rv = (fx < 0.5f) ? 1 : -1;
				break;
			case sawtooth:
				rv = fx;
				break;
			case inverse_sawtooth:
				rv = 1.0f - fx;
				break;
			default:
				rv = 0;
				break;
		}

		return rv*m_amplitude+m_base;
	}


	TCMod_rotate::
	TCMod_rotate(const float speed) : m_speed(speed*0.001) {
	}

	void
	TCMod_rotate::
	apply() {
		glRotatef(m_speed*Loop::current_time(), 0, 0, 1);
		CHECK_OPENGL_ERROR;
	}

	TCMod_scale::
	TCMod_scale(const float hscale,
		    const float vscale) : m_hscale(hscale),
					  m_vscale(vscale) {
	}

	void
	TCMod_scale::
	apply() {
		glTranslatef(m_hscale, m_vscale, 1);
		CHECK_OPENGL_ERROR;
	}
	
	TCMod_stretch::
	TCMod_stretch(Waveform* waveform) : m_waveform(waveform) {
	}

	void
	TCMod_stretch::
	apply() {
		float scale = m_waveform->value(Loop::current_time()*0.001);
		glScalef(scale, scale, 0);
		CHECK_OPENGL_ERROR;
	}
	
	TCMod_translate::
	TCMod_translate(const float hspeed,
			const float vspeed) : m_hspeed(hspeed*0.001),
					      m_vspeed(-vspeed*0.001) {
	}

	void
	TCMod_translate::
	apply() {
		glTranslatef(m_hspeed*Loop::current_time(),
			     m_vspeed*Loop::current_time(), 0);
		CHECK_OPENGL_ERROR;
	}
	
	TCMod_turb::
	TCMod_turb(const float amplitude,
		   const float phase,
		   const float frequency) : m_amplitude(amplitude),
					    m_phase(phase),
					    m_frequency(frequency) {
	}

	void
	TCMod_turb::
	apply() {
		float angle = (Loop::current_time()*0.001+m_phase)*m_frequency*2*M_PI;
		glTranslatef(m_amplitude*cos(angle), m_amplitude*sin(angle), 0);
		CHECK_OPENGL_ERROR;
	}
	

	

	void
	SimpleShader::
	draw(VertexArray* vertex_array, Topology* topology,
	     Texture& texture, Texture& lightmap) {
		glDepthMask(GL_TRUE);
		glDepthFunc(GL_LEQUAL);
		glDisable(GL_BLEND);
		OpenGL::cull_face(GL_FRONT);
		OpenGL::alpha_func(GL_ALWAYS);
		texture.bind();
		Texture::set_texture_unit(1);
		lightmap.bind();
		Texture::set_texture_unit(0);
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glColor4f(1.0, 1.0, 1.0, 1.0);

		vertex_array->glPointer();
		topology->draw();
		CHECK_OPENGL_ERROR;
	}

	ComplexShader::
	ComplexShader() : m_cullface(GL_FRONT) {
	}

	bool
	ComplexShader::
	add_layer(Layer* l) {
		bool rv = m_layers.empty();

		m_layers.push_back(l);

		return rv;
	}

	void
	ComplexShader::
	draw(VertexArray* vertex_array, Topology* topology,
	     Texture& texture, Texture& lightmap) {
		glDepthFunc(GL_LEQUAL);

		Texture::set_texture_unit(1);
		Texture::none().bind();
		Texture::set_texture_unit(0);

		OpenGL::cull_face(m_cullface);

		list<Layer*>::iterator i;
		bool first = true;

		for (i = m_layers.begin(); i != m_layers.end(); ++i) {
			if (first) {
				if (m_blend) {
					glDepthMask(GL_FALSE);
					glEnable(GL_BLEND);
				} else {
					glDepthMask(GL_TRUE);
					glDisable(GL_BLEND);
				}
				first = false;
			} else {
				glDepthMask(GL_FALSE);
				glEnable(GL_BLEND);
			}

			bool texmatrix_modified = (*i)->bind(texture, lightmap);

			if ((*i)->m_tcgen == Layer::tc_lightmap) {
				vertex_array->glPointer_lightmap();
			} else if ((*i)->m_tcgen == Layer::tc_texture) {
				vertex_array->glPointer_texture();
			} else if ((*i)->m_tcgen == Layer::tc_environment) {
				Texture::tc_gen(GL_SPHERE_MAP);
				vertex_array->glPointer_texture();
			} else {
				vertex_array->glPointer_texture();
			}
			topology->draw();

			if ((*i)->m_tcgen == Layer::tc_environment) {
				Texture::tc_gen(GL_NONE);
			}

			if (texmatrix_modified) {
				glMatrixMode(GL_TEXTURE);
				glLoadIdentity();
				glMatrixMode(GL_MODELVIEW);
			}
		}
		glDepthMask(GL_TRUE);
		CHECK_OPENGL_ERROR;
	}

	ComplexShader::Layer::
	Layer() : m_texture(no_texture),
		  m_tcgen(tc_texture),
		  m_layertexture(),
		  m_sfactor(GL_ONE),
		  m_dfactor(GL_ZERO),
		  m_alphafunc(GL_ALWAYS),
		  m_alpharef(0.0),
		  m_rgbgen(NULL),
		  m_alphagen(NULL) {
	}

	void
	ComplexShader::Layer::
	add_tcmod(TCMod* tcmod) {
		m_tcmods.push_back(tcmod);
	}

	bool
	ComplexShader::Layer::
	bind(Texture& t, Texture& l) {
		float rgb = 1.0;
		float alpha = 1.0;

		switch (m_texture) {
			case no_texture:
				Texture::none().bind();
				break;
			case texture:
				t.bind();
				break;
			case lightmap:
				l.bind();
				break;
			case layertexture:
				m_layertexture.bind();
				break;
		}

		OpenGL::alpha_func(m_alphafunc, m_alpharef);

		glBlendFunc(m_sfactor, m_dfactor);

		if (m_rgbgen) {
			rgb = m_rgbgen->value(Loop::current_time()*0.001);
		}

		if (m_alphagen) {
			alpha = m_alphagen->value(Loop::current_time()*0.001);
		}

		glColor4f(rgb, rgb, rgb, alpha);

		CHECK_OPENGL_ERROR;

		if (!m_tcmods.empty()) {
			list<TCMod*>::iterator i;

			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();

			glTranslatef(0.5, 0.5, 0);
			for (i = m_tcmods.begin(); i != m_tcmods.end(); ++i) {
				(*i)->apply();
			}
			glTranslatef(-0.5, -0.5, 0);
		
			glMatrixMode(GL_MODELVIEW);

			CHECK_OPENGL_ERROR;

			return true;
		}
		return false;
	}

	Shader::
	Shader() : m_shader(),
		   m_texture(Texture::none()),
		   m_lightmap(Texture::none()) {
	}

	Shader::
	Shader(const Shader& src) : m_shader(src.m_shader),
				    m_texture(src.m_texture),
				    m_lightmap(src.m_lightmap) {
	}

	Shader::
	Shader(BaseShader* shader,
	       const Texture& texture,
	       const Texture& lightmap) : m_shader(shader),
					  m_texture(texture),
					  m_lightmap(lightmap) {
	}

	Shader&
	Shader::
	operator=(const Shader& src) {
		m_shader = src.m_shader;
		m_texture = src.m_texture;
		m_lightmap = src.m_lightmap;
		return *this;
	}

	void
	Shader::
	set_shader(BaseShader* shader) {
		m_shader = shader;
	}

	void
	Shader::
	set_texture(const Texture& texture) {
		m_texture = texture;
	}

	void
	Shader::
	set_lightmap(const Texture& lightmap) {
		m_lightmap = lightmap;
	}

}

