/**************************************************************************

    This file is part of "The Science and Magick 3d Engine", yet
    another OpenGL-based engine.
    Copyright (C) 2001 - 2002 Rodolphe Suescun

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
    $Id: shader.h,v 1.4 2003/08/31 10:04:34 rodzilla Exp $
*/

#ifndef __SAM3D_SHADER_H__
#define __SAM3D_SHADER_H__

#include <SaM3d/opengl.h>
#include <SaM3d/texture.h>
#include <list>

namespace SaM3d {

	class Topology;
	class VertexArray;


	class Waveform {
	public:
		enum waveform_t { sine, triangle, square,
				  sawtooth, inverse_sawtooth };

	protected:
		waveform_t m_waveform;
		float m_base;
		float m_amplitude;
		float m_phase;
		float m_frequency;

	public:
		Waveform(const waveform_t waveform,
			 const float base,
			 const float amplitude,
			 const float phase,
			 const float frequency);

		float value(const float x);
	};

	class TCMod {
	public:
		virtual ~TCMod() { }
		virtual void apply() = 0;
	};
	
	class TCMod_rotate :  public TCMod {
		float m_speed;

	public:
		TCMod_rotate(const float speed);
		virtual void apply();
	};

	class TCMod_scale :  public TCMod {
		float m_hscale;
		float m_vscale;

	public:
		TCMod_scale(const float hscale, const float vscale);
		virtual void apply();
	};

	class TCMod_stretch :  public TCMod {
		Waveform* m_waveform;

	public:
		TCMod_stretch(Waveform* waveform);
		virtual void apply();
	};

	class TCMod_translate :  public TCMod {
		float m_hspeed;
		float m_vspeed;

	public:
		TCMod_translate(const float hspeed, const float vspeed);
		virtual void apply();
	};

	class TCMod_turb :  public TCMod {
		float m_amplitude;
		float m_phase;
		float m_frequency;

	public:
		TCMod_turb(const float amplitude, const float phase, const float frequency);
		virtual void apply();
	};



	class BaseShader : public RefCounted {
	public:
		bool	m_blend;

		BaseShader() : m_blend(false) { }

		virtual void draw(VertexArray* vertex_array,
				  Topology* topology,
				  Texture& texture, Texture& lightmap) { }
	};

	class SimpleShader : public BaseShader {
	public:
		virtual void draw(VertexArray* vertex_array,
				  Topology* topology,
				  Texture& texture, Texture& lightmap);
	};

	class ComplexShader : public BaseShader {
	public:
		class Layer {
			std::list<TCMod*>	m_tcmods;
		public:
			enum {
				no_texture, texture, lightmap, layertexture
			} 	m_texture;
			enum {
				tc_texture, tc_lightmap, tc_environment
			} 		m_tcgen;
			Texture		m_layertexture;
			GLenum		m_sfactor;
			GLenum		m_dfactor;
			GLenum		m_alphafunc;
			float		m_alpharef;
			Waveform*	m_rgbgen;
			Waveform*	m_alphagen;

			Layer();

			void add_tcmod(TCMod* tcmod);

			bool bind(Texture& t, Texture& l);
		};
	protected:
		std::list<Layer*>	m_layers;

	public:
		GLenum m_cullface;

		ComplexShader();

		bool add_layer(Layer* layer);

		virtual void draw(VertexArray* vertex_array,
				  Topology* topology,
				  Texture& texture, Texture& lightmap);
	};

	class Shader {
		SmartPtr<BaseShader>	m_shader;
		Texture			m_texture;
		Texture			m_lightmap;

	public:
		Shader();
		Shader(const Shader& src);
		Shader(BaseShader* shader,
		       const Texture& texture, const Texture& lightmap);

		Shader& operator=(const Shader& src);

		void set_shader(BaseShader* shader);
		void set_texture(const Texture& texture);
		void set_lightmap(const Texture& lightmap);

		void draw(VertexArray* vertex_array, Topology* topology) {
			m_shader->draw(vertex_array, topology, m_texture, m_lightmap);
		}

		bool is_transparent() { return m_shader->m_blend; }
	};

}

#endif
